"""
 @file
 @brief Unit tests for ColorGrade editor helpers
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import unittest

PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from windows.color_grade_editor import (  # noqa: E402
    _set_color_value,
    _set_keyframe_value,
    _default_curve_node,
    WheelRow,
    colorgrade_keyframe_frames,
    curve_enabled_at_frame,
    curve_nodes_at_frame,
    default_curve_data,
    default_wheels_data,
    is_achromatic_color,
    normalize_curve_data,
    normalize_wheels_data,
    puck_display_color,
    wheel_snapshot,
    wheels_enabled_at_frame,
)
from windows.models.properties_model import PropertiesModel  # noqa: E402
from qt_api import QColor  # noqa: E402
import openshot  # noqa: E402


class ColorGradeEditorTests(unittest.TestCase):
    def test_default_curve_data_uses_linear_nodes(self):
        curve = default_curve_data()
        self.assertEqual(len(curve["nodes"]), 2)
        self.assertEqual(curve["nodes"][0]["interpolation"], openshot.LINEAR)
        self.assertEqual(curve["nodes"][1]["interpolation"], openshot.LINEAR)
        self.assertEqual(curve["nodes"][0]["x"]["Points"][0]["interpolation"], openshot.LINEAR)

    def test_new_curve_node_interpolates_after_frame_one_shape_is_added(self):
        node = _default_curve_node(2, 0.5, 0.8, frame_number=24)
        node["y"] = _set_keyframe_value(node["y"], 1, 0.2)
        curve = normalize_curve_data({
            "enabled": {"Points": [{"co": {"X": 1.0, "Y": 1.0}, "interpolation": openshot.LINEAR}]},
            "nodes": [
                _default_curve_node(0, 0.0, 0.0),
                node,
                _default_curve_node(1, 1.0, 1.0),
            ],
        })

        evaluated = {node["id"]: node for node in curve_nodes_at_frame(curve, 12)}

        self.assertGreater(evaluated[2]["y"], 0.2)
        self.assertLess(evaluated[2]["y"], 0.8)

    def test_normalize_curve_data_falls_back_to_default(self):
        self.assertEqual(normalize_curve_data({}), default_curve_data())

    def test_curve_helpers_evaluate_current_frame(self):
        curve = default_curve_data()
        curve["nodes"][0]["y"]["Points"][0]["co"]["Y"] = 0.0
        curve["nodes"][0]["y"]["Points"].append({"co": {"X": 10.0, "Y": 0.5}, "interpolation": openshot.LINEAR})
        nodes = curve_nodes_at_frame(curve, 10)
        self.assertAlmostEqual(nodes[0]["y"], 0.5, places=3)
        self.assertTrue(curve_enabled_at_frame(curve, 1))

    def test_set_keyframe_value_preserves_other_frames(self):
        keyframe = {"Points": [
            {"co": {"X": 1.0, "Y": 0.1}, "interpolation": openshot.LINEAR},
            {"co": {"X": 20.0, "Y": 0.9}, "interpolation": openshot.LINEAR},
        ]}
        updated = _set_keyframe_value(keyframe, 10, 0.5)
        self.assertEqual(len(updated["Points"]), 3)
        self.assertAlmostEqual(updated["Points"][0]["co"]["Y"], 0.1, places=3)
        self.assertAlmostEqual(updated["Points"][1]["co"]["Y"], 0.5, places=3)
        self.assertAlmostEqual(updated["Points"][2]["co"]["Y"], 0.9, places=3)

    def test_set_color_value_preserves_other_frames(self):
        color = {
            "red": {"Points": [
                {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                {"co": {"X": 20.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
            ]},
            "green": {"Points": [
                {"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
                {"co": {"X": 20.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
            ]},
            "blue": {"Points": [
                {"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
                {"co": {"X": 20.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
            ]},
            "alpha": {"Points": [
                {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                {"co": {"X": 20.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
            ]},
        }
        updated = _set_color_value(color, 10, QColor("#ffffff"))
        self.assertEqual(len(updated["red"]["Points"]), 3)
        self.assertAlmostEqual(updated["red"]["Points"][0]["co"]["Y"], 255.0, places=3)
        self.assertAlmostEqual(updated["red"]["Points"][1]["co"]["Y"], 255.0, places=3)
        self.assertAlmostEqual(updated["red"]["Points"][2]["co"]["Y"], 0.0, places=3)
        self.assertEqual(wheel_snapshot({"color": updated, "amount": {"Points": [{"co": {"X": 10.0, "Y": 1.0}, "interpolation": openshot.LINEAR}]}, "luma": 0.0}, 10)["color"], "#ffffff")

    def test_normalize_wheels_data_clamps_values(self):
        wheels = normalize_wheels_data({
            "global": {"color": "#zzzzzz", "amount": 5, "luma": -5},
        })
        snapshot = wheel_snapshot(wheels["global"], 1)
        self.assertEqual(snapshot["color"], "#ffffff")
        self.assertEqual(snapshot["amount"], 1.0)
        self.assertEqual(snapshot["luma"], -1.0)

    def test_normalize_wheels_data_supplies_missing_entries(self):
        wheels = normalize_wheels_data({})
        self.assertEqual(wheels, default_wheels_data())

    def test_normalize_wheels_data_preserves_enabled_flag(self):
        wheels = normalize_wheels_data({"enabled": False})
        self.assertFalse(wheels_enabled_at_frame(wheels, 1))

    def test_colorgrade_keyframe_frames_includes_wheel_subkeyframes(self):
        wheels = default_wheels_data()
        wheels["global"]["color_keyframes"]["red"]["Points"].append({
            "co": {"X": 24.0, "Y": 64.0},
            "interpolation": openshot.CONSTANT,
        })
        wheels["highlights"]["luma_keyframes"]["Points"].append({
            "co": {"X": 48.0, "Y": 0.2},
            "interpolation": openshot.LINEAR,
        })

        self.assertEqual(
            colorgrade_keyframe_frames(wheels, "colorgrade_wheels"),
            {1, 24, 48},
        )

    def test_properties_model_applies_interpolation_to_colorgrade_wheel_frame(self):
        wheels = default_wheels_data()
        for keyframe in (
            wheels["global"]["color_keyframes"]["red"],
            wheels["global"]["color_keyframes"]["green"],
            wheels["global"]["amount_keyframes"],
            wheels["shadows"]["luma_keyframes"],
        ):
            keyframe["Points"].append({
                "co": {"X": 24.0, "Y": 0.5},
                "interpolation": openshot.LINEAR,
            })

        helper = PropertiesModel.__new__(PropertiesModel)
        updated, changed = helper._apply_colorgrade_interpolation(
            wheels,
            "colorgrade_wheels",
            1,
            24,
            openshot.CONSTANT,
            [],
        )

        self.assertTrue(changed)
        self.assertEqual(updated["global"]["color_keyframes"]["red"]["Points"][1]["interpolation"], openshot.CONSTANT)
        self.assertEqual(updated["global"]["color_keyframes"]["green"]["Points"][1]["interpolation"], openshot.CONSTANT)
        self.assertEqual(updated["global"]["amount_keyframes"]["Points"][1]["interpolation"], openshot.CONSTANT)
        self.assertEqual(updated["shadows"]["luma_keyframes"]["Points"][1]["interpolation"], openshot.CONSTANT)

    def test_properties_model_applies_interpolation_to_colorgrade_curve_frame(self):
        curve = default_curve_data()
        curve["nodes"][0]["y"]["Points"].append({
            "co": {"X": 32.0, "Y": 0.4},
            "interpolation": openshot.LINEAR,
        })
        curve["nodes"][1]["x"]["Points"].append({
            "co": {"X": 32.0, "Y": 0.8},
            "interpolation": openshot.LINEAR,
        })

        helper = PropertiesModel.__new__(PropertiesModel)
        updated, changed = helper._apply_colorgrade_interpolation(
            curve,
            "colorgrade_curve",
            1,
            32,
            openshot.CONSTANT,
            [],
        )

        self.assertTrue(changed)
        self.assertEqual(updated["nodes"][0]["y"]["Points"][1]["interpolation"], openshot.CONSTANT)
        self.assertEqual(updated["nodes"][1]["x"]["Points"][1]["interpolation"], openshot.CONSTANT)

    def test_properties_model_inserts_colorgrade_wheel_keyframe_without_resetting_payload(self):
        wheels = default_wheels_data()
        wheels["global"]["amount_keyframes"]["Points"].append({
            "co": {"X": 24.0, "Y": 0.8},
            "interpolation": openshot.LINEAR,
        })
        wheels["shadows"]["luma_keyframes"]["Points"].append({
            "co": {"X": 24.0, "Y": -0.4},
            "interpolation": openshot.LINEAR,
        })

        helper = PropertiesModel.__new__(PropertiesModel)
        updated = helper._insert_colorgrade_keyframe(wheels, "colorgrade_wheels", 12)

        self.assertIn(12, colorgrade_keyframe_frames(updated, "colorgrade_wheels"))
        self.assertIn(24, colorgrade_keyframe_frames(updated, "colorgrade_wheels"))
        self.assertEqual(updated["global"]["amount_keyframes"]["Points"][1]["interpolation"], openshot.LINEAR)
        self.assertEqual(updated["global"]["amount_keyframes"]["Points"][-1]["co"]["X"], 24.0)
        self.assertEqual(updated["shadows"]["luma_keyframes"]["Points"][-1]["co"]["X"], 24.0)

    def test_properties_model_removes_colorgrade_wheel_frame_without_clearing_other_frames(self):
        wheels = default_wheels_data()
        for keyframe in (
            wheels["global"]["color_keyframes"]["red"],
            wheels["global"]["color_keyframes"]["green"],
            wheels["global"]["amount_keyframes"],
            wheels["highlights"]["luma_keyframes"],
        ):
            keyframe["Points"].append({
                "co": {"X": 24.0, "Y": 0.5},
                "interpolation": openshot.LINEAR,
            })

        helper = PropertiesModel.__new__(PropertiesModel)
        updated, changed = helper._remove_colorgrade_keyframe(wheels, "colorgrade_wheels", 24)

        self.assertTrue(changed)
        self.assertEqual(colorgrade_keyframe_frames(updated, "colorgrade_wheels"), {1})

    def test_wheel_row_remove_keyframe_targets_active_interpolated_frame(self):
        row = WheelRow.__new__(WheelRow)
        row._frame_number = 12
        row._data = normalize_wheels_data({
            "global": {
                "color_keyframes": {
                    "red": {"Points": [
                        {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                        {"co": {"X": 24.0, "Y": 64.0}, "interpolation": openshot.LINEAR},
                    ]},
                    "green": {"Points": [
                        {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                        {"co": {"X": 24.0, "Y": 128.0}, "interpolation": openshot.LINEAR},
                    ]},
                    "blue": {"Points": [
                        {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                        {"co": {"X": 24.0, "Y": 192.0}, "interpolation": openshot.LINEAR},
                    ]},
                    "alpha": {"Points": [
                        {"co": {"X": 1.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                        {"co": {"X": 24.0, "Y": 255.0}, "interpolation": openshot.LINEAR},
                    ]},
                },
                "amount_keyframes": {"Points": [
                    {"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
                    {"co": {"X": 24.0, "Y": 0.8}, "interpolation": openshot.LINEAR},
                ]},
            }
        })["global"]
        row.dragStarted = type("Signal", (), {"emit": lambda self: None})()
        row.dragFinished = type("Signal", (), {"emit": lambda self: None})()
        row.changed = type("Signal", (), {"emit": lambda self: None})()
        row._apply_data = lambda: None

        row._remove_keyframe()

        self.assertEqual(row._frame_set(), {1})

    def test_wheel_row_remove_slider_keyframe_targets_active_interpolated_frame(self):
        row = WheelRow.__new__(WheelRow)
        row._frame_number = 12
        row._data = normalize_wheels_data({
            "global": {
                "amount_keyframes": {"Points": [
                    {"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
                    {"co": {"X": 24.0, "Y": 0.8}, "interpolation": openshot.LINEAR},
                ]},
                "luma_keyframes": {"Points": [
                    {"co": {"X": 1.0, "Y": 0.0}, "interpolation": openshot.LINEAR},
                    {"co": {"X": 24.0, "Y": -0.4}, "interpolation": openshot.LINEAR},
                ]},
            }
        })["global"]
        row.dragStarted = type("Signal", (), {"emit": lambda self: None})()
        row.dragFinished = type("Signal", (), {"emit": lambda self: None})()
        row.changed = type("Signal", (), {"emit": lambda self: None})()
        row._apply_data = lambda: None

        row._remove_slider_keyframe("amount")

        self.assertEqual(
            [point["co"]["X"] for point in row._data["amount_keyframes"]["Points"]],
            [1.0],
        )
        self.assertEqual(
            [point["co"]["X"] for point in row._data["luma_keyframes"]["Points"]],
            [1.0, 24.0],
        )

    def test_achromatic_color_detection_treats_white_as_neutral(self):
        self.assertTrue(is_achromatic_color(QColor("#ffffff")))
        self.assertTrue(is_achromatic_color(QColor("#808080")))
        self.assertFalse(is_achromatic_color(QColor("#00ff24")))

    def test_puck_display_color_blends_from_neutral_to_hue(self):
        neutral = puck_display_color({"color": "#ff0000", "amount": 0.0})
        full = puck_display_color({"color": "#ff0000", "amount": 1.0})
        self.assertNotEqual(neutral.name(), full.name())
        self.assertEqual(full.name(), "#ff0000")


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Reusable widgets for the Recording dock.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 SPDX-License-Identifier: GPL-3.0-or-later
 """

import ctypes
import ctypes.util
import os
import re
import shutil
import subprocess  # nosec B404 -- fixed argv only; shell execution is never used
import sys

from qt_api import (
    Qt, pyqtSignal, QRect, QPoint, QEvent, QTimer,
    QWidget, QFrame, QLabel, QPushButton, QVBoxLayout, QHBoxLayout,
    QGridLayout, QApplication, QDialog, QPainter, QColor, QPen,
)

from classes.app import get_app
from classes.logger import log


_WIN32_USER32 = None
_WIN32_DWMAPI = None
_WIN32_ENUMPROC = None
_WIN32_MONITOR_ENUMPROC = None
_WIN32_TYPES = None
_MAC_COREGRAPHICS = None
_MAC_COREFOUNDATION = None
_MAC_KEYS = None


class RecordingSourceCard(QFrame):
    toggled = pyqtSignal(bool)

    def __init__(self, title, subtitle, symbol, parent=None):
        super().__init__(parent)
        self._checked = False
        self._available = True
        self.setObjectName("recordingCard")
        self.setProperty("checked", False)
        self.setProperty("available", True)
        self.setCursor(Qt.PointingHandCursor)
        self.setMinimumHeight(88)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(4)

        icon = QLabel(symbol, self)
        icon.setObjectName("recordingCardIcon")
        icon.setAlignment(Qt.AlignCenter)
        icon.setFixedHeight(30)
        layout.addWidget(icon, 0, Qt.AlignCenter)

        title_label = QLabel(title, self)
        title_label.setObjectName("recordingCardTitle")
        title_label.setAlignment(Qt.AlignCenter)
        subtitle_label = QLabel(subtitle, self)
        subtitle_label.setObjectName("recordingCardSubtitle")
        subtitle_label.setAlignment(Qt.AlignCenter)
        subtitle_label.setWordWrap(True)
        layout.addWidget(title_label)
        layout.addWidget(subtitle_label)

    def isChecked(self):
        return self._checked

    def setChecked(self, checked):
        checked = bool(checked) and self._available
        if self._checked == checked:
            return
        self._checked = checked
        self.setProperty("checked", checked)
        self.style().unpolish(self)
        self.style().polish(self)
        self.toggled.emit(checked)

    def setAvailable(self, available, tooltip=""):
        self._available = bool(available)
        self.setEnabled(self._available)
        self.setProperty("available", self._available)
        self.setToolTip(tooltip)
        if not self._available:
            self.setChecked(False)
        self.style().unpolish(self)
        self.style().polish(self)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self._available:
            self.setChecked(not self._checked)
        super().mousePressEvent(event)


class RecordingSection(QFrame):
    def __init__(self, title, symbol, parent=None):
        super().__init__(parent)
        self.setObjectName("recordingSection")
        self.setProperty("active", True)
        self.outer_layout = QVBoxLayout(self)
        self.outer_layout.setContentsMargins(10, 8, 10, 10)
        self.outer_layout.setSpacing(8)

        header = QHBoxLayout()
        icon = QLabel(symbol, self)
        icon.setObjectName("recordingSectionIcon")
        title_label = QLabel(title, self)
        title_label.setObjectName("recordingSectionTitle")
        self.advanced_button = QPushButton(get_app()._tr("Advanced"), self)
        self.advanced_button.setObjectName("recordingAdvancedLink")
        self.advanced_button.setFlat(True)
        self.advanced_button.setCheckable(True)
        self.advanced_button.setCursor(Qt.PointingHandCursor)
        header.addWidget(icon)
        header.addWidget(title_label)
        header.addStretch()
        header.addWidget(self.advanced_button)
        self.outer_layout.addLayout(header)

        self.body = QWidget(self)
        self.body_layout = QVBoxLayout(self.body)
        self.body_layout.setContentsMargins(0, 0, 0, 0)
        self.body_layout.setSpacing(8)
        self.outer_layout.addWidget(self.body)

        self.advanced_body = QWidget(self)
        self.advanced_layout = QGridLayout(self.advanced_body)
        self.advanced_layout.setContentsMargins(0, 0, 0, 0)
        self.advanced_layout.setHorizontalSpacing(8)
        self.advanced_layout.setVerticalSpacing(5)
        self.advanced_body.hide()
        self.outer_layout.addWidget(self.advanced_body)
        self.advanced_button.toggled.connect(self.advanced_body.setVisible)

    def setActive(self, active):
        active = bool(active)
        self.setProperty("active", active)
        self.body.setVisible(active)
        self.advanced_button.setEnabled(active)
        if not active:
            self.advanced_body.hide()
            self.advanced_button.setChecked(False)
        self.style().unpolish(self)
        self.style().polish(self)


class SegmentButton(QPushButton):
    def __init__(self, text, parent=None):
        super().__init__(text, parent)
        self.setObjectName("recordingSegment")
        self.setCheckable(True)
        self.setCursor(Qt.PointingHandCursor)


class RegionSelectorOverlay(QDialog):
    def __init__(self, parent=None, screen_geometry=None, x11_geometry=True):
        super().__init__(parent)
        self.origin = QPoint()
        self.selection = QRect()
        self.global_origin = QPoint()
        self.global_selection = QRect()
        self.screen_geometry = screen_geometry
        self.x11_geometry = bool(x11_geometry)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.Tool)
        self.setMouseTracking(True)
        self.setCursor(Qt.CrossCursor)
        self.setWindowOpacity(0.35)

    def select(self):
        screen_geometry = self.screen_geometry
        if screen_geometry is None:
            screen = QApplication.primaryScreen()
            screen_geometry = screen.geometry() if screen else None
        if screen_geometry:
            self.setGeometry(screen_geometry)
        accepted = getattr(QDialog, "Accepted", None)
        if accepted is None and hasattr(QDialog, "DialogCode"):
            accepted = QDialog.DialogCode.Accepted
        if self.exec_() == accepted and self.selection.isValid():
            rect = self.selection.normalized()
            if self.x11_geometry:
                return self._to_x11_geometry(rect, screen_geometry)
            return self._to_global_geometry(rect)
        return None

    def _activate_for_keyboard(self):
        """Ensure modal picker keyboard shortcuts reach this window."""
        self.raise_()
        self.activateWindow()
        self.setFocus(Qt.ActiveWindowFocusReason)

    def eventFilter(self, watched, event):
        # The Windows tool window can briefly lose focus while its parent is
        # hidden. Catch Escape anywhere in this application while the modal
        # picker is active, rather than depending on child/window focus.
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Escape:
            self.reject()
            return True
        return super().eventFilter(watched, event)

    def showEvent(self, event):
        super().showEvent(event)
        QApplication.instance().installEventFilter(self)
        QTimer.singleShot(0, self._activate_for_keyboard)

    def done(self, result):
        app = QApplication.instance()
        if app:
            app.removeEventFilter(self)
        super().done(result)

    def _to_global_geometry(self, rect):
        top_left = self.mapToGlobal(rect.topLeft())
        bottom_right = self.mapToGlobal(rect.bottomRight())
        x = int(min(top_left.x(), bottom_right.x()))
        y = int(min(top_left.y(), bottom_right.y()))
        width = int(abs(bottom_right.x() - top_left.x()) + 1)
        height = int(abs(bottom_right.y() - top_left.y()) + 1)
        return x, y, max(1, width), max(1, height)

    def _to_x11_geometry(self, rect, screen_geometry):
        root_width, root_height = x11_root_size()
        top_left = self.mapToGlobal(rect.topLeft())
        bottom_right = self.mapToGlobal(rect.bottomRight())
        if screen_geometry and screen_geometry.width() > 0 and screen_geometry.height() > 0:
            scale_x = float(root_width or screen_geometry.width()) / float(screen_geometry.width())
            scale_y = float(root_height or screen_geometry.height()) / float(screen_geometry.height())
            origin_x = screen_geometry.x()
            origin_y = screen_geometry.y()
        else:
            scale_x = scale_y = 1.0
            origin_x = origin_y = 0

        x = int(round((top_left.x() - origin_x) * scale_x))
        y = int(round((top_left.y() - origin_y) * scale_y))
        width = int(round((bottom_right.x() - top_left.x() + 1) * scale_x))
        height = int(round((bottom_right.y() - top_left.y() + 1) * scale_y))

        if root_width and root_height:
            x = max(0, min(x, root_width - 1))
            y = max(0, min(y, root_height - 1))
            width = max(1, min(width, root_width - x))
            height = max(1, min(height, root_height - y))
        return x, y, width, height

    def _global_pos(self, event):
        if hasattr(event, "globalPosition"):
            return event.globalPosition().toPoint()
        return event.globalPos()

    def mousePressEvent(self, event):
        self.origin = event.pos()
        self.selection = QRect(self.origin, self.origin)
        self.global_origin = self._global_pos(event)
        self.global_selection = QRect(self.global_origin, self.global_origin)
        self.update()

    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton:
            self.selection = QRect(self.origin, event.pos())
            self.global_selection = QRect(self.global_origin, self._global_pos(event))
            self.update()

    def mouseReleaseEvent(self, event):
        self.selection = QRect(self.origin, event.pos()).normalized()
        self.global_selection = QRect(self.global_origin, self._global_pos(event)).normalized()
        if self.global_selection.width() > 8 and self.global_selection.height() > 8:
            self.accept()
        else:
            self.reject()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.reject()
        super().keyPressEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), QColor(0, 0, 0, 120))
        if self.global_selection.isValid():
            global_rect = self.global_selection.normalized()
            rect = QRect(
                self.mapFromGlobal(global_rect.topLeft()),
                self.mapFromGlobal(global_rect.bottomRight()),
            ).normalized()
            painter.setPen(QPen(QColor(64, 145, 255), 2))
            painter.fillRect(rect, QColor(64, 145, 255, 45))
            painter.drawRect(rect)


def pick_x11_region(parent=None):
    return RegionSelectorOverlay(parent).select()


def pick_screen_region(parent=None, capture_geometry=None, overlay_geometry=None):
    if sys.platform.startswith("win"):
        return pick_windows_region(parent, capture_geometry, overlay_geometry)
    if sys.platform == "darwin":
        return pick_mac_region(parent)
    return pick_x11_region(parent)


def x11_root_size():
    try:
        result = subprocess.run(  # nosec B603 -- fixed argv list, no shell
            ["xwininfo", "-root"],
            check=True,
            text=True,
            capture_output=True,
            timeout=5,
        )
    except Exception:
        return None, None

    match = re.search(r"Width:\s+(\d+).*?Height:\s+(\d+)", result.stdout, re.S)
    if match:
        return int(match.group(1)), int(match.group(2))
    match = re.search(r"Dimensions:\s+(\d+)x(\d+)", result.stdout)
    if match:
        return int(match.group(1)), int(match.group(2))
    return None, None


def screen_root_size():
    if sys.platform.startswith("win"):
        _, _, width, height = windows_virtual_screen_geometry()
        return width, height
    if sys.platform == "darwin":
        _, _, width, height = mac_primary_screen_geometry()
        return width, height
    return x11_root_size()


def screen_root_geometry():
    if sys.platform.startswith("win"):
        return windows_virtual_screen_geometry()
    if sys.platform == "darwin":
        return mac_primary_screen_geometry()
    width, height = x11_root_size()
    return 0, 0, width, height


def screen_capture_sources():
    if sys.platform.startswith("win"):
        sources = windows_monitor_geometries()
        all_x, all_y, all_width, all_height = windows_virtual_screen_geometry()
        if all_width and all_height and len(sources) > 1:
            sources.append({
                "id": "all",
                "label": get_app()._tr("All Screens"),
                "display": "desktop",
                "x": int(all_x or 0),
                "y": int(all_y or 0),
                "width": int(all_width),
                "height": int(all_height),
                "all": True,
                "primary": False,
            })
        return sources

    screens = list(QApplication.screens() or [])
    if sys.platform == "darwin":
        sources = []
        primary = QApplication.primaryScreen()
        for index, screen in enumerate(screens):
            geometry = screen.geometry()
            scale = float(screen.devicePixelRatio() or 1.0)
            label = screen.name() or get_app()._tr("Screen %s") % (index + 1)
            sources.append({
                "id": "screen-%s" % index,
                "label": "%s (%sx%s)" % (
                    label,
                    int(round(geometry.width() * scale)),
                    int(round(geometry.height() * scale)),
                ),
                "display": "Capture screen %s:none" % index,
                "x": 0,
                "y": 0,
                "width": max(1, int(round(geometry.width() * scale))),
                "height": max(1, int(round(geometry.height() * scale))),
                "all": False,
                "primary": screen is primary,
            })
        if sources:
            return sources
        x, y, width, height = mac_primary_screen_geometry()
        return [{
            "id": "screen-0",
            "label": get_app()._tr("Screen 1"),
            "display": "Capture screen 0:none",
            "x": int(x or 0),
            "y": int(y or 0),
            "width": int(width or 1280),
            "height": int(height or 720),
            "all": False,
            "primary": True,
        }]

    display = os.environ.get("DISPLAY", ":0.0")
    root_width, root_height = x11_root_size()
    virtual = None
    try:
        primary = QApplication.primaryScreen()
        virtual = primary.virtualGeometry() if primary else None
    except Exception:
        virtual = None
    scale_x = 1.0
    scale_y = 1.0
    if virtual and virtual.width() > 0 and virtual.height() > 0 and root_width and root_height:
        scale_x = float(root_width) / float(virtual.width())
        scale_y = float(root_height) / float(virtual.height())
    sources = []
    primary = QApplication.primaryScreen()
    for index, screen in enumerate(screens):
        geometry = screen.geometry()
        origin_x = virtual.x() if virtual else 0
        origin_y = virtual.y() if virtual else 0
        x = int(round((geometry.x() - origin_x) * scale_x))
        y = int(round((geometry.y() - origin_y) * scale_y))
        width = max(1, int(round(geometry.width() * scale_x)))
        height = max(1, int(round(geometry.height() * scale_y)))
        label = screen.name() or get_app()._tr("Screen %s") % (index + 1)
        sources.append({
            "id": "screen-%s" % index,
            "label": "%s (%sx%s)" % (label, width, height),
            "display": display,
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "all": False,
            "primary": screen is primary,
        })
    if root_width and root_height and len(sources) > 1:
        sources.append({
            "id": "all",
            "label": get_app()._tr("All Screens"),
            "display": display,
            "x": 0,
            "y": 0,
            "width": int(root_width),
            "height": int(root_height),
            "all": True,
            "primary": False,
        })
    if sources:
        return sources
    return [{
        "id": "screen-0",
        "label": get_app()._tr("Screen 1"),
        "display": display,
        "x": 0,
        "y": 0,
        "width": int(root_width or 1280),
        "height": int(root_height or 720),
        "all": False,
        "primary": True,
    }]


def pick_x11_window():
    result = pick_x11_window_with_xdotool()
    if result:
        return result
    return pick_x11_window_with_xwininfo()


def pick_screen_window():
    if sys.platform.startswith("win"):
        return pick_windows_window()
    if sys.platform == "darwin":
        return pick_mac_window()
    return pick_x11_window()


def mac_primary_screen_geometry():
    if sys.platform != "darwin":
        return 0, 0, None, None
    try:
        screen = QApplication.primaryScreen()
        geometry = screen.geometry() if screen else None
        if not geometry:
            return 0, 0, None, None
        scale = float(screen.devicePixelRatio() or 1.0)
        return (
            0,
            0,
            max(1, int(round(geometry.width() * scale))),
            max(1, int(round(geometry.height() * scale))),
        )
    except Exception:
        log.debug("Unable to query macOS primary screen geometry", exc_info=True)
        return 0, 0, None, None


def _mac_point_to_capture_pixels(x, y, width, height):
    screen = QApplication.screenAt(QPoint(int(x), int(y))) or QApplication.primaryScreen()
    geometry = screen.geometry() if screen else None
    scale = float(screen.devicePixelRatio() or 1.0) if screen else 1.0
    origin_x = geometry.x() if geometry else 0
    origin_y = geometry.y() if geometry else 0
    return (
        max(0, int(round((int(x) - origin_x) * scale))),
        max(0, int(round((int(y) - origin_y) * scale))),
        max(1, int(round(int(width) * scale))),
        max(1, int(round(int(height) * scale))),
    )


def pick_mac_region(parent=None):
    screen = QApplication.primaryScreen()
    geometry = screen.geometry() if screen else None
    result = RegionSelectorOverlay(parent, geometry, x11_geometry=False).select()
    if not result:
        return None
    return _mac_point_to_capture_pixels(*result)


def _mac_frameworks():
    global _MAC_COREGRAPHICS, _MAC_COREFOUNDATION
    if _MAC_COREGRAPHICS is None:
        path = ctypes.util.find_library("CoreGraphics")
        _MAC_COREGRAPHICS = ctypes.cdll.LoadLibrary(path) if path else False
    if _MAC_COREFOUNDATION is None:
        path = ctypes.util.find_library("CoreFoundation")
        _MAC_COREFOUNDATION = ctypes.cdll.LoadLibrary(path) if path else False
    if not _MAC_COREGRAPHICS or not _MAC_COREFOUNDATION:
        return None, None
    return _MAC_COREGRAPHICS, _MAC_COREFOUNDATION


def _mac_define_signatures():
    CG, CF = _mac_frameworks()
    if not CG or not CF:
        return None, None

    CFArrayRef = ctypes.c_void_p
    CFTypeRef = ctypes.c_void_p
    CFStringRef = ctypes.c_void_p
    CFNumberRef = ctypes.c_void_p
    CFIndex = ctypes.c_long

    CG.CGWindowListCopyWindowInfo.argtypes = [ctypes.c_uint32, ctypes.c_uint32]
    CG.CGWindowListCopyWindowInfo.restype = CFArrayRef

    CF.CFRelease.argtypes = [CFTypeRef]
    CF.CFRelease.restype = None
    CF.CFArrayGetCount.argtypes = [CFArrayRef]
    CF.CFArrayGetCount.restype = CFIndex
    CF.CFArrayGetValueAtIndex.argtypes = [CFArrayRef, CFIndex]
    CF.CFArrayGetValueAtIndex.restype = CFTypeRef
    CF.CFDictionaryGetValueIfPresent.argtypes = [ctypes.c_void_p, CFTypeRef, ctypes.POINTER(CFTypeRef)]
    CF.CFDictionaryGetValueIfPresent.restype = ctypes.c_bool
    CF.CFStringCreateWithCString.argtypes = [CFTypeRef, ctypes.c_char_p, ctypes.c_uint32]
    CF.CFStringCreateWithCString.restype = CFStringRef
    CF.CFStringGetCString.argtypes = [CFStringRef, ctypes.c_char_p, CFIndex, ctypes.c_uint32]
    CF.CFStringGetCString.restype = ctypes.c_bool
    CF.CFNumberGetValue.argtypes = [CFNumberRef, ctypes.c_int, ctypes.c_void_p]
    CF.CFNumberGetValue.restype = ctypes.c_bool
    return CG, CF


def _mac_keys():
    global _MAC_KEYS
    if _MAC_KEYS is not None:
        return _MAC_KEYS
    _CG, CF = _mac_define_signatures()
    if not CF:
        return {}
    encoding = 0x08000100
    names = (
        "kCGWindowNumber",
        "kCGWindowOwnerName",
        "kCGWindowOwnerPID",
        "kCGWindowName",
        "kCGWindowLayer",
        "kCGWindowAlpha",
        "kCGWindowBounds",
        "X",
        "Y",
        "Width",
        "Height",
    )
    _MAC_KEYS = {
        name: CF.CFStringCreateWithCString(None, name.encode("utf-8"), encoding)
        for name in names
    }
    return _MAC_KEYS


def _mac_dict_value(dictionary, key):
    _CG, CF = _mac_define_signatures()
    keys = _mac_keys()
    if not CF or key not in keys:
        return None
    value = ctypes.c_void_p()
    if CF.CFDictionaryGetValueIfPresent(dictionary, keys[key], ctypes.byref(value)):
        return value.value
    return None


def _mac_number(value, as_int=False):
    if not value:
        return None
    _CG, CF = _mac_define_signatures()
    if not CF:
        return None
    if as_int:
        out = ctypes.c_int()
        return int(out.value) if CF.CFNumberGetValue(value, 3, ctypes.byref(out)) else None
    out = ctypes.c_double()
    return float(out.value) if CF.CFNumberGetValue(value, 13, ctypes.byref(out)) else None


def _mac_string(value):
    if not value:
        return ""
    _CG, CF = _mac_define_signatures()
    if not CF:
        return ""
    buffer = ctypes.create_string_buffer(4096)
    if CF.CFStringGetCString(value, buffer, len(buffer), 0x08000100):
        return buffer.value.decode("utf-8", "replace")
    return ""


def _mac_window_bounds(bounds):
    if not bounds:
        return None
    x = _mac_number(_mac_dict_value(bounds, "X"))
    y = _mac_number(_mac_dict_value(bounds, "Y"))
    width = _mac_number(_mac_dict_value(bounds, "Width"))
    height = _mac_number(_mac_dict_value(bounds, "Height"))
    if x is None or y is None or width is None or height is None:
        return None
    return int(round(x)), int(round(y)), int(round(width)), int(round(height))


def _mac_visible_windows():
    CG, CF = _mac_define_signatures()
    if not CG or not CF:
        return []
    options = (1 << 0) | (1 << 4)  # On-screen only, excluding desktop elements.
    window_array = CG.CGWindowListCopyWindowInfo(options, 0)
    if not window_array:
        return []
    windows = []
    try:
        count = int(CF.CFArrayGetCount(window_array))
        current_pid = os.getpid()
        for index in range(count):
            window = CF.CFArrayGetValueAtIndex(window_array, index)
            layer = _mac_number(_mac_dict_value(window, "kCGWindowLayer"), as_int=True)
            alpha = _mac_number(_mac_dict_value(window, "kCGWindowAlpha"))
            if layer != 0 or alpha == 0:
                continue
            owner = _mac_string(_mac_dict_value(window, "kCGWindowOwnerName"))
            owner_pid = _mac_number(_mac_dict_value(window, "kCGWindowOwnerPID"), as_int=True)
            title = _mac_string(_mac_dict_value(window, "kCGWindowName"))
            bounds = _mac_window_bounds(_mac_dict_value(window, "kCGWindowBounds"))
            number = _mac_number(_mac_dict_value(window, "kCGWindowNumber"), as_int=True)
            if not bounds:
                continue
            x, y, width, height = bounds
            if width <= 8 or height <= 8:
                continue
            if owner_pid == current_pid:
                continue
            windows.append({
                "id": number or 0,
                "owner": owner,
                "title": title,
                "x": x,
                "y": y,
                "width": width,
                "height": height,
            })
    finally:
        CF.CFRelease(window_array)
    return windows


def _mac_pick_window_at(x, y):
    for window in _mac_visible_windows():
        left = window["x"]
        top = window["y"]
        right = left + window["width"]
        bottom = top + window["height"]
        if left <= x <= right and top <= y <= bottom:
            return window
    return None


class MacWindowSelectorOverlay(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        screen = QApplication.primaryScreen()
        self.screen_geometry = screen.geometry() if screen else QRect(0, 0, 1, 1)
        self.hover_rect = QRect()
        self.selected_result = None
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.Tool)
        self.setMouseTracking(True)
        self.setCursor(Qt.CrossCursor)
        self.setWindowOpacity(0.22)
        self.setGeometry(self.screen_geometry)

    def select(self):
        accepted = getattr(QDialog, "Accepted", None)
        if accepted is None and hasattr(QDialog, "DialogCode"):
            accepted = QDialog.DialogCode.Accepted
        return self.selected_result if self.exec_() == accepted else None

    def _event_global_pos(self, event):
        if hasattr(event, "globalPosition"):
            return event.globalPosition().toPoint()
        return event.globalPos()

    def _update_hover(self, point):
        picked = _mac_pick_window_at(int(point.x()), int(point.y()))
        if not picked:
            self.hover_rect = QRect()
            self.update()
            return None
        self.hover_rect = QRect(
            self.mapFromGlobal(QPoint(picked["x"], picked["y"])),
            self.mapFromGlobal(QPoint(picked["x"] + picked["width"], picked["y"] + picked["height"])),
        ).normalized()
        self.update()
        return picked

    def mouseMoveEvent(self, event):
        self._update_hover(self._event_global_pos(event))

    def mouseReleaseEvent(self, event):
        picked = self._update_hover(self._event_global_pos(event))
        if not picked:
            self.reject()
            return
        x, y, width, height = _mac_point_to_capture_pixels(
            picked["x"], picked["y"], picked["width"], picked["height"])
        if width > 8 and height > 8:
            log.info(
                "Selected macOS window: id=%s owner=%s title=%s x=%s y=%s width=%s height=%s",
                picked["id"], picked["owner"], picked["title"], x, y, width, height,
            )
            self.selected_result = x, y, width, height, str(picked["id"])
            self.accept()
        else:
            self.reject()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.reject()
        super().keyPressEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), QColor(0, 0, 0, 95))
        if self.hover_rect.isValid():
            painter.setPen(QPen(QColor(64, 145, 255), 3))
            painter.fillRect(self.hover_rect, QColor(64, 145, 255, 45))
            painter.drawRect(self.hover_rect)


def pick_mac_window(parent=None):
    try:
        return MacWindowSelectorOverlay(parent).select()
    except Exception as ex:
        log.warning("Unable to select macOS window: %s", ex, exc_info=True)
        return None


def _windows_enum_proc_type():
    global _WIN32_ENUMPROC
    if _WIN32_ENUMPROC is None:
        wintypes = _windows_types()
        _WIN32_ENUMPROC = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)
    return _WIN32_ENUMPROC


def _windows_monitor_enum_proc_type():
    global _WIN32_MONITOR_ENUMPROC
    if _WIN32_MONITOR_ENUMPROC is None:
        wintypes = _windows_types()
        _WIN32_MONITOR_ENUMPROC = ctypes.WINFUNCTYPE(
            wintypes.BOOL,
            wintypes.HMONITOR,
            wintypes.HDC,
            ctypes.POINTER(wintypes.RECT),
            wintypes.LPARAM,
        )
    return _WIN32_MONITOR_ENUMPROC


def _windows_types():
    global _WIN32_TYPES
    if _WIN32_TYPES is None:
        from ctypes import wintypes
        _WIN32_TYPES = wintypes
    return _WIN32_TYPES


def _windows_user32():
    global _WIN32_USER32
    if _WIN32_USER32 is None:
        wintypes = _windows_types()
        user32 = ctypes.windll.user32
        user32.GetSystemMetrics.argtypes = [ctypes.c_int]
        user32.GetSystemMetrics.restype = ctypes.c_int
        user32.GetWindowTextLengthW.argtypes = [wintypes.HWND]
        user32.GetWindowTextLengthW.restype = ctypes.c_int
        user32.GetWindowTextW.argtypes = [wintypes.HWND, wintypes.LPWSTR, ctypes.c_int]
        user32.GetWindowTextW.restype = ctypes.c_int
        user32.GetWindowRect.argtypes = [wintypes.HWND, ctypes.POINTER(wintypes.RECT)]
        user32.GetWindowRect.restype = wintypes.BOOL
        user32.IsWindowVisible.argtypes = [wintypes.HWND]
        user32.IsWindowVisible.restype = wintypes.BOOL
        user32.IsIconic.argtypes = [wintypes.HWND]
        user32.IsIconic.restype = wintypes.BOOL
        user32.GetWindowThreadProcessId.argtypes = [wintypes.HWND, ctypes.POINTER(wintypes.DWORD)]
        user32.GetWindowThreadProcessId.restype = wintypes.DWORD
        user32.EnumWindows.argtypes = [_windows_enum_proc_type(), wintypes.LPARAM]
        user32.EnumWindows.restype = wintypes.BOOL
        user32.EnumDisplayMonitors.argtypes = [
            wintypes.HDC,
            ctypes.POINTER(wintypes.RECT),
            _windows_monitor_enum_proc_type(),
            wintypes.LPARAM,
        ]
        user32.EnumDisplayMonitors.restype = wintypes.BOOL
        user32.GetCursorPos.argtypes = [ctypes.POINTER(wintypes.POINT)]
        user32.GetCursorPos.restype = wintypes.BOOL
        _WIN32_USER32 = user32
    return _WIN32_USER32


def _windows_dwmapi():
    global _WIN32_DWMAPI
    if _WIN32_DWMAPI is None:
        wintypes = _windows_types()
        try:
            dwmapi = ctypes.windll.dwmapi
            dwmapi.DwmGetWindowAttribute.argtypes = [
                wintypes.HWND, ctypes.c_uint, ctypes.c_void_p, ctypes.c_uint,
            ]
            dwmapi.DwmGetWindowAttribute.restype = ctypes.c_long
            _WIN32_DWMAPI = dwmapi
        except Exception:
            _WIN32_DWMAPI = False
    return _WIN32_DWMAPI or None


def windows_virtual_screen_geometry():
    if not sys.platform.startswith("win"):
        return 0, 0, None, None
    user32 = _windows_user32()
    x = int(user32.GetSystemMetrics(76))      # SM_XVIRTUALSCREEN
    y = int(user32.GetSystemMetrics(77))      # SM_YVIRTUALSCREEN
    width = int(user32.GetSystemMetrics(78))  # SM_CXVIRTUALSCREEN
    height = int(user32.GetSystemMetrics(79)) # SM_CYVIRTUALSCREEN
    return x, y, width, height


def windows_monitor_geometries():
    if not sys.platform.startswith("win"):
        return []
    user32 = _windows_user32()
    sources = []

    def enum_proc(_hmonitor, _hdc, rect, _lparam):
        geometry = rect.contents
        width = int(geometry.right - geometry.left)
        height = int(geometry.bottom - geometry.top)
        if width > 0 and height > 0:
            index = len(sources) + 1
            sources.append({
                "id": "screen-%s" % index,
                "label": "%s (%sx%s)" % (get_app()._tr("Screen %s") % index, width, height),
                "display": "desktop",
                "x": int(geometry.left),
                "y": int(geometry.top),
                "width": width,
                "height": height,
                "all": False,
                "primary": int(geometry.left) <= 0 < int(geometry.right)
                    and int(geometry.top) <= 0 < int(geometry.bottom),
            })
        return True

    try:
        user32.EnumDisplayMonitors(None, None, _windows_monitor_enum_proc_type()(enum_proc), 0)
    except Exception:
        log.debug("Unable to enumerate Windows monitors", exc_info=True)
        return []
    return sources


class WindowsRegionSelectorOverlay(RegionSelectorOverlay):
    """Select a gdigrab rectangle in native Windows desktop pixels."""

    def __init__(self, parent=None, capture_geometry=None, overlay_geometry=None):
        x, y, width, height = capture_geometry or windows_virtual_screen_geometry()
        self.capture_geometry = QRect(int(x), int(y), int(width or 1), int(height or 1))
        super().__init__(
            parent,
            overlay_geometry or self.capture_geometry,
            x11_geometry=False,
        )
        self.native_origin = None
        self.native_selection = QRect()

    def _native_cursor_pos(self):
        position = _windows_cursor_pos()
        return QPoint(*position) if position is not None else None

    def _clamped_native_point(self, point):
        bounds = self.capture_geometry
        return QPoint(
            max(bounds.left(), min(point.x(), bounds.right())),
            max(bounds.top(), min(point.y(), bounds.bottom())),
        )

    def mousePressEvent(self, event):
        point = self._native_cursor_pos()
        if point is None:
            return super().mousePressEvent(event)
        self.native_origin = self._clamped_native_point(point)
        self.native_selection = QRect(self.native_origin, self.native_origin)
        self.update()

    def mouseMoveEvent(self, event):
        if self.native_origin is None or not (event.buttons() & Qt.LeftButton):
            return
        point = self._native_cursor_pos()
        if point is not None:
            self.native_selection = QRect(
                self.native_origin, self._clamped_native_point(point)).normalized()
            self.update()

    def mouseReleaseEvent(self, event):
        if self.native_origin is None:
            return super().mouseReleaseEvent(event)
        point = self._native_cursor_pos()
        if point is not None:
            self.native_selection = QRect(
                self.native_origin, self._clamped_native_point(point)).normalized()
        if self.native_selection.width() > 8 and self.native_selection.height() > 8:
            self.accept()
        else:
            self.reject()

    def selected_geometry(self):
        rect = self.native_selection.normalized()
        return rect.x(), rect.y(), max(1, rect.width()), max(1, rect.height())

    def select(self):
        accepted = getattr(QDialog, "Accepted", None)
        if accepted is None and hasattr(QDialog, "DialogCode"):
            accepted = QDialog.DialogCode.Accepted
        self.setGeometry(self.screen_geometry)
        return self.selected_geometry() if self.exec_() == accepted else None

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), QColor(0, 0, 0, 120))
        if self.native_selection.isValid():
            bounds = self.capture_geometry
            scale_x = float(self.width() or 1) / float(bounds.width() or 1)
            scale_y = float(self.height() or 1) / float(bounds.height() or 1)
            native = self.native_selection.normalized()
            rect = QRect(
                int(round((native.x() - bounds.x()) * scale_x)),
                int(round((native.y() - bounds.y()) * scale_y)),
                max(1, int(round(native.width() * scale_x))),
                max(1, int(round(native.height() * scale_y))),
            )
            painter.setPen(QPen(QColor(64, 145, 255), 2))
            painter.fillRect(rect, QColor(64, 145, 255, 45))
            painter.drawRect(rect)


def _windows_region_overlay_geometry(capture_geometry):
    """Find the Qt logical screen rect corresponding to native capture bounds."""
    x, y, width, height = capture_geometry
    candidates = []
    for screen in QApplication.screens() or []:
        geometry = screen.geometry()
        scale = float(screen.devicePixelRatio() or 1.0)
        size_error = (
            abs(int(round(geometry.width() * scale)) - int(width))
            + abs(int(round(geometry.height() * scale)) - int(height))
        )
        origin_match = geometry.contains(QPoint(int(x), int(y)))
        candidates.append((0 if origin_match else 1, size_error, geometry))
    if candidates:
        return min(candidates, key=lambda candidate: candidate[:2])[2]
    return QRect(int(x), int(y), int(width), int(height))


def _windows_physical_point_to_qt(x, y):
    """Map a native desktop pixel to Qt's logical desktop coordinates."""
    for monitor in windows_monitor_geometries():
        left = int(monitor["x"])
        top = int(monitor["y"])
        width = int(monitor["width"])
        height = int(monitor["height"])
        # Window rectangles use exclusive right/bottom edges, so accept the
        # outer desktop boundary as well as ordinary points inside a monitor.
        if left <= x <= left + width and top <= y <= top + height:
            logical = _windows_region_overlay_geometry(
                (left, top, width, height))
            return QPoint(
                logical.x() + int(round((x - left) * logical.width() / width)),
                logical.y() + int(round((y - top) * logical.height() / height)),
            )
    return QPoint(int(x), int(y))


def pick_windows_region(parent=None, capture_geometry=None, overlay_geometry=None):
    uses_virtual_desktop = capture_geometry is None
    x, y, width, height = capture_geometry or windows_virtual_screen_geometry()
    if not width or not height:
        return None
    if overlay_geometry is None:
        if uses_virtual_desktop:
            primary = QApplication.primaryScreen()
            overlay_geometry = primary.virtualGeometry() if primary else None
        overlay_geometry = overlay_geometry or _windows_region_overlay_geometry(
            (x, y, width, height))
    return WindowsRegionSelectorOverlay(
        parent,
        (x, y, width, height),
        overlay_geometry,
    ).select()


def _windows_window_title(hwnd):
    user32 = _windows_user32()
    length = user32.GetWindowTextLengthW(hwnd)
    if length <= 0:
        return ""
    buffer = ctypes.create_unicode_buffer(length + 1)
    user32.GetWindowTextW(hwnd, buffer, length + 1)
    return buffer.value


def _windows_hwnd_value(hwnd):
    value = getattr(hwnd, "value", hwnd)
    return int(value or 0)


def _windows_window_rect(hwnd):
    wintypes = _windows_types()
    user32 = _windows_user32()
    rect = wintypes.RECT()
    dwmapi = _windows_dwmapi()
    if dwmapi is not None:
        # DWMWA_EXTENDED_FRAME_BOUNDS gives visual bounds without the invisible
        # resize border on modern Windows.
        if dwmapi.DwmGetWindowAttribute(hwnd, 9, ctypes.byref(rect), ctypes.sizeof(rect)) == 0:
            return rect
    if user32.GetWindowRect(hwnd, ctypes.byref(rect)):
        return rect
    return None


def _windows_cursor_pos():
    wintypes = _windows_types()
    point = wintypes.POINT()
    if _windows_user32().GetCursorPos(ctypes.byref(point)):
        return int(point.x), int(point.y)
    return None


def _windows_pick_window_at(x, y, excluded_hwnd=None):
    wintypes = _windows_types()
    user32 = _windows_user32()
    excluded_hwnd = int(excluded_hwnd or 0)
    point = wintypes.POINT(int(x), int(y))
    candidates = []

    def enum_proc(hwnd, _lparam):
        if not user32.IsWindowVisible(hwnd) or user32.IsIconic(hwnd):
            return True
        # The selector itself covers the desktop and is the first top-level
        # hit. Do not exclude the entire current process: users must be able
        # to select OpenShot's main window for recording.
        if excluded_hwnd and _windows_hwnd_value(hwnd) == excluded_hwnd:
            return True
        title = _windows_window_title(hwnd)
        if not title:
            return True
        rect = _windows_window_rect(hwnd)
        if not rect:
            return True
        width = int(rect.right - rect.left)
        height = int(rect.bottom - rect.top)
        if width <= 8 or height <= 8:
            return True
        if rect.left <= point.x <= rect.right and rect.top <= point.y <= rect.bottom:
            candidates.append((hwnd, rect, title))
            return False
        return True

    user32.EnumWindows(_windows_enum_proc_type()(enum_proc), 0)
    return candidates[0] if candidates else None


class WindowsWindowSelectorOverlay(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        x, y, width, height = windows_virtual_screen_geometry()
        self.capture_geometry = QRect(int(x), int(y), int(width or 1), int(height or 1))
        primary = QApplication.primaryScreen()
        self.screen_geometry = (
            primary.virtualGeometry() if primary else self.capture_geometry)
        self.hover_rect = QRect()
        self.hover_title = ""
        self.selected_result = None
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint | Qt.Tool)
        self.setMouseTracking(True)
        self.setCursor(Qt.CrossCursor)
        self.setWindowOpacity(0.22)
        self.setGeometry(self.screen_geometry)

    def select(self):
        accepted = getattr(QDialog, "Accepted", None)
        if accepted is None and hasattr(QDialog, "DialogCode"):
            accepted = QDialog.DialogCode.Accepted
        return self.selected_result if self.exec_() == accepted else None

    def _activate_for_keyboard(self):
        self.raise_()
        self.activateWindow()
        self.setFocus(Qt.ActiveWindowFocusReason)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Escape:
            self.reject()
            return True
        return super().eventFilter(watched, event)

    def showEvent(self, event):
        super().showEvent(event)
        QApplication.instance().installEventFilter(self)
        QTimer.singleShot(0, self._activate_for_keyboard)

    def done(self, result):
        app = QApplication.instance()
        if app:
            app.removeEventFilter(self)
        super().done(result)

    def _physical_to_overlay_point(self, x, y):
        global_point = _windows_physical_point_to_qt(int(x), int(y))
        return self.mapFromGlobal(global_point)

    def _update_hover(self):
        cursor_pos = _windows_cursor_pos()
        if cursor_pos is None:
            return None
        picked = _windows_pick_window_at(
            cursor_pos[0], cursor_pos[1], _windows_hwnd_value(self.winId()))
        if not picked:
            self.hover_rect = QRect()
            self.hover_title = ""
            self.update()
            return None
        _hwnd, rect, title = picked
        self.hover_rect = QRect(
            self._physical_to_overlay_point(rect.left, rect.top),
            self._physical_to_overlay_point(rect.right, rect.bottom),
        ).normalized()
        self.hover_title = title
        self.update()
        return picked

    def mouseMoveEvent(self, event):
        self._update_hover()

    def mouseReleaseEvent(self, event):
        picked = self._update_hover()
        if not picked:
            self.reject()
            return
        hwnd, rect, title = picked
        width = int(rect.right - rect.left)
        height = int(rect.bottom - rect.top)
        if width > 8 and height > 8:
            hwnd_value = _windows_hwnd_value(hwnd)
            log.info(
                "Selected Windows window: hwnd=%s title=%s x=%s y=%s width=%s height=%s",
                hwnd_value, title, int(rect.left), int(rect.top), width, height,
            )
            self.selected_result = int(rect.left), int(rect.top), width, height, str(hwnd_value)
            self.accept()
        else:
            self.reject()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.reject()
        super().keyPressEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(self.rect(), QColor(0, 0, 0, 95))
        if self.hover_rect.isValid():
            painter.setPen(QPen(QColor(64, 145, 255), 3))
            painter.fillRect(self.hover_rect, QColor(64, 145, 255, 45))
            painter.drawRect(self.hover_rect)


def pick_windows_window(parent=None):
    overlay = WindowsWindowSelectorOverlay(parent)
    return overlay.select()


def pick_x11_window_with_xdotool():
    if not shutil.which("xdotool"):
        return None
    try:
        selected = subprocess.run(  # nosec B603 -- fixed argv list, no shell
            ["xdotool", "selectwindow"],
            check=True,
            text=True,
            capture_output=True,
            timeout=30,
        )
        window_id = selected.stdout.strip()
        if not re.fullmatch(r"(?:0[xX][0-9a-fA-F]+|[0-9]+)", window_id):
            return None
        geometry = subprocess.run(  # nosec B603 -- validated ID, argv list, no shell
            ["xdotool", "getwindowgeometry", "--shell", window_id],
            check=True,
            text=True,
            capture_output=True,
            timeout=5,
        )
    except Exception as ex:
        log.debug("Unable to select X11 window with xdotool: %s", ex, exc_info=True)
        return None

    parsed = {}
    for line in geometry.stdout.splitlines():
        if "=" not in line:
            continue
        key, value = line.split("=", 1)
        if key in ("X", "Y", "WIDTH", "HEIGHT"):
            try:
                parsed[key.lower()] = int(value)
            except ValueError:
                return None
    if not all(key in parsed for key in ("x", "y", "width", "height")):
        return None
    log.info(
        "Selected X11 window with xdotool: id=%s x=%s y=%s width=%s height=%s",
        window_id, parsed["x"], parsed["y"], parsed["width"], parsed["height"],
    )
    return parsed["x"], parsed["y"], parsed["width"], parsed["height"], window_id


def pick_x11_window_with_xwininfo():
    try:
        result = subprocess.run(  # nosec B603 -- fixed argv list, no shell
            ["xwininfo"],
            check=True,
            text=True,
            capture_output=True,
            timeout=30,
        )
    except Exception as ex:
        log.warning("Unable to select X11 window with xwininfo: %s", ex)
        return None

    fields = {
        "x": r"Absolute upper-left X:\s+(-?\d+)",
        "y": r"Absolute upper-left Y:\s+(-?\d+)",
        "width": r"Width:\s+(\d+)",
        "height": r"Height:\s+(\d+)",
        "border": r"Border width:\s+(\d+)",
    }
    parsed = {}
    for key, pattern in fields.items():
        match = re.search(pattern, result.stdout)
        if not match and key != "border":
            return None
        parsed[key] = int(match.group(1)) if match else 0
    # xwininfo reports width/height for the window interior, but X11 border
    # pixels can be outside that rectangle. Capture the interior by offsetting
    # the origin inward while preserving the reported content size.
    parsed["x"] += parsed["border"]
    parsed["y"] += parsed["border"]
    log.info(
        "Selected X11 window with xwininfo: x=%s y=%s width=%s height=%s border=%s",
        parsed["x"], parsed["y"], parsed["width"], parsed["height"], parsed["border"],
    )
    return parsed["x"], parsed["y"], parsed["width"], parsed["height"], ""



"""
 @file
 @brief This file loads the clip cutting interface (quickly cut up a clip into smaller clips)
  @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import functools
import json

from qt_api import pyqtSignal, QTimer, QSize
from qt_api import Qt, QEvent
from qt_api import QIcon
from qt_api import QDialog, QMessageBox, QSizePolicy, QSlider, QToolButton, QLineEdit
import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes import info, ui_util, time_parts
from classes.app import get_app
from classes.clip_utils import is_single_image_media
from classes.logger import log
from classes.metrics import track_metric_screen
from classes.proxy_service import dialog_preview_reader_data
from windows.preview_thread import PreviewParent
from windows.video_widget import VideoWidget


class Cutting(QDialog):
    """ Cutting Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'cutting.ui')

    # Signals for preview thread
    previewFrameSignal = pyqtSignal(int)
    refreshFrameSignal = pyqtSignal()
    LoadFileSignal = pyqtSignal(str)
    PlaySignal = pyqtSignal()
    PauseSignal = pyqtSignal()
    SeekSignal = pyqtSignal(int)
    LoadTimelineAndSeekSignal = pyqtSignal(int)
    SpeedSignal = pyqtSignal(float)
    StopSignal = pyqtSignal()

    @staticmethod
    def _preview_window_title(file_obj, translate):
        _ = translate
        if not file_obj or not hasattr(file_obj, "data") or not isinstance(file_obj.data, dict):
            return _("Preview")
        friendly_name = str(file_obj.data.get("name") or "").strip()
        if not friendly_name:
            friendly_name = os.path.basename(str(file_obj.data.get("path") or "").strip())
        if friendly_name:
            return _("Preview: %s") % friendly_name
        return _("Preview")

    def __init__(self, file=None, preview=False):
        _ = get_app()._tr
        self.is_preview_mode = preview
        self._preview_autoplay_active = preview
        self._preview_autoplay_attempts = 0
        self._shutdown_in_progress = False
        self._close_after_shutdown = False
        self.loop_playback = bool(preview)

        # Create dialog class
        super().__init__()

        # Load UI from designer
        ui_util.load_ui(self, self.ui_path)

        # Init UI
        ui_util.init_ui(self)
        self.setWindowFlags(
            (self.windowFlags() & ~Qt.Dialog)
            | Qt.Window
            | Qt.WindowMinMaxButtonsHint
            | Qt.WindowMaximizeButtonHint
        )
        self.setSizeGripEnabled(True)

        # Track metrics
        track_metric_screen("cutting-screen")

        # Keep track of file object
        self.file = file
        self.source_reader_data = dialog_preview_reader_data(file, prefer_proxy=False)
        self.proxy_reader_data = dialog_preview_reader_data(file, prefer_proxy=True)
        self.reader_data = self.proxy_reader_data
        self.file_path = str(self.reader_data.get("path") or file.absolute_path() or "")
        self.video_length = int(file.data['video_length'])
        self.fps_num = int(file.data['fps']['num'])
        self.fps_den = int(file.data['fps']['den'])
        self.fps = float(self.fps_num) / float(self.fps_den)
        self.width = int(file.data['width'])
        self.height = int(file.data['height'])
        self.sample_rate = int(get_app().project.get("sample_rate"))
        self.channels = int(file.data['channels'])
        self.channel_layout = int(file.data['channel_layout'])

        self.start_frame = 1
        self.start_image = None
        self.end_frame = self.video_length
        self.end_image = None

        # If preview, hide cutting controls
        if preview:
            self.lblInstructions.setVisible(False)
            self.widgetControls.setVisible(False)
            self.setWindowTitle(self._preview_window_title(file, _))

        self.previous_start = 0.0
        if float(file.data.get("start", 0.0)) > 0.0:
            self.start_frame = round(file.data.get("start", 0) * self.fps) + 1

            # Remember the previous start property (on init)
            self.previous_start = self.file.data.get("start", 0.0)
        if float(file.data.get("end", 0.0)) > 0.0:
            self.end_frame = round(file.data.get("end", 0) * self.fps)
            self.video_length = (self.end_frame - self.start_frame) + 1

        # Set clip start / end
        clip_start = file.data.get("start", 0.0)
        clip_end = file.data.get("end", file.data.get("duration", 0.0))

        # Open video file with Reader
        log.info(self.file_path)

        # Add Video Widget
        self.videoPreview = VideoWidget(watch_project=False)
        self.videoPreview.win = self
        self.videoPreview.setObjectName("videoPreview")
        self.videoPreview.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.verticalLayout.insertWidget(0, self.videoPreview)

        # Set max size of video preview (for speed)
        viewport_rect = self.videoPreview.centeredViewport(self.videoPreview.width(), self.videoPreview.height())

        try:
            self._build_preview_timeline(self.reader_data, QSize(viewport_rect.width(), viewport_rect.height()))
        except Exception:
            log.error(
                'Failed to load media file into preview player: %s',
                self.file_path)
            return

        # Start the preview thread
        self.initialized = False
        self.transforming_clip = False
        self.preview_parent = PreviewParent()
        self.preview_parent.Init(self, self.r, self.videoPreview, self.video_length)
        self.preview_thread = self.preview_parent.worker

        # Set slider constraints
        self.sliderIgnoreSignal = False
        self.sliderVideo.setMinimum(1)
        self.sliderVideo.setMaximum(self.video_length)
        self.sliderVideo.setSingleStep(1)
        self.sliderVideo.setPageStep(24)
        if self.is_preview_mode:
            self._build_preview_repeat_button()

        # Initialize first frame display.
        # For cutting mode, preserve the legacy two-step seek refresh.
        # For preview mode, avoid the seek/pause startup hack so autoplay
        # isn't fighting initialization pauses.
        if self.is_preview_mode:
            self.sliderIgnoreSignal = True
            self.sliderVideo.setValue(1)
            self.sliderIgnoreSignal = False
        else:
            QTimer.singleShot(500, functools.partial(self.sliderVideo.setValue, 2))
            QTimer.singleShot(600, functools.partial(self.sliderVideo.setValue, 1))

        # Connect signals
        self.actionPlay.triggered.connect(self.actionPlay_Triggered)
        self.btnPlay.clicked.connect(self.btnPlay_clicked)
        self.sliderVideo.valueChanged.connect(self.sliderVideo_valueChanged)
        self.sliderVideo.sliderReleased.connect(self.sliderVideo_released)
        self.btnStart.clicked.connect(self.btnStart_clicked)
        self.btnEnd.clicked.connect(self.btnEnd_clicked)
        self.btnClear.clicked.connect(self.btnClear_clicked)
        self.btnAddClip.clicked.connect(self.btnAddClip_clicked)
        self.txtName.installEventFilter(self)
        self.sliderVideo.installEventFilter(self)
        # Timer to ensure final preview update
        self.slider_timer = QTimer(self)
        self.slider_timer.setInterval(100)
        self.slider_timer.setSingleShot(True)
        self.slider_timer.timeout.connect(self.sliderVideo_timeout)
        self.videoPreview.delayed_resize_timer.timeout.connect(self._apply_dynamic_preview_max_size)
        self.initialized = True

    def _build_preview_repeat_button(self):
        _ = get_app()._tr
        self.btnRepeat = QToolButton(self)
        self.btnRepeat.setObjectName("btnRepeat")
        self.btnRepeat.setCheckable(True)
        self.btnRepeat.setChecked(True)
        self.btnRepeat.setAutoRaise(True)
        self.btnRepeat.setFixedSize(24, 24)
        self.btnRepeat.setToolTip(_("Repeat"))
        self.btnRepeat.setStyleSheet(
            "QToolButton#btnRepeat { border-radius: 4px; }"
            "QToolButton#btnRepeat:checked { background-color: rgba(83,160,237,80); }"
        )
        self.btnRepeat.toggled.connect(self._on_repeat_toggled)
        self.horizontalLayout_3.insertWidget(2, self.btnRepeat)

        icon = ui_util.get_icon("media-playlist-repeat")
        if icon is None or icon.isNull():
            icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", "tool-media-repeat.svg")
            icon = QIcon(icon_path)
        self.btnRepeat.setIcon(icon)

    def _on_repeat_toggled(self, checked):
        self.loop_playback = bool(checked)

    def keyPressEvent(self, event):
        if event and event.key() == Qt.Key_Space:
            focused = self.focusWidget()
            if focused and isinstance(focused, QLineEdit):
                return super(Cutting, self).keyPressEvent(event)
            if hasattr(self, "btnPlay") and self.btnPlay is not None:
                self.btnPlay.click()
                event.accept()
                return
        return super(Cutting, self).keyPressEvent(event)

    def _target_preview_max_size(self):
        viewport_rect = self.videoPreview.centeredViewport(self.videoPreview.width(), self.videoPreview.height())
        device_pixel_ratio = self.devicePixelRatioF()
        requested = QSize(
            max(2, int(round(viewport_rect.width() * device_pixel_ratio))),
            max(2, int(round(viewport_rect.height() * device_pixel_ratio))),
        )

        source_width = int(getattr(self, "width", 0) or 0)
        source_height = int(getattr(self, "height", 0) or 0)
        file_data = getattr(getattr(self, "file", None), "data", {})
        if is_single_image_media(getattr(self, "reader_data", {})) or is_single_image_media(file_data):
            project = getattr(get_app(), "project", None)
            if project:
                source_width = max(source_width, int(project.get("width") or 0))
                source_height = max(source_height, int(project.get("height") or 0))
        if source_width > 0 and source_height > 0:
            source_size = QSize(source_width, source_height)
            if requested.width() > source_width or requested.height() > source_height:
                capped = QSize(source_size)
            else:
                capped = QSize(requested)
        else:
            capped = requested

        if capped.height() > 0:
            ratio = float(capped.width()) / float(capped.height())
            even_width = max(2, int(round(capped.width() / 2.0) * 2))
            even_height = max(2, int(round(round(even_width / ratio) / 2.0) * 2))
            capped = QSize(even_width, even_height)

        return capped

    def _reader_capacity(self, reader_data):
        try:
            return int(reader_data.get("width", 0) or 0), int(reader_data.get("height", 0) or 0)
        except Exception:
            return 0, 0

    def _select_reader_data_for_size(self, target_size):
        _ = target_size
        proxy_data = getattr(self, "proxy_reader_data", None) or {}
        source_data = getattr(self, "source_reader_data", None) or {}
        proxy_path = str(proxy_data.get("path") or "")
        if proxy_path:
            return proxy_data
        return source_data or proxy_data

    def _build_preview_timeline(self, reader_data, max_size):
        self.reader_data = reader_data
        self.file_path = str(reader_data.get("path") or self.file.absolute_path() or "")
        is_single_image = bool(is_single_image_media(reader_data) or is_single_image_media(getattr(self.file, "data", {})))
        source_path = str(getattr(self.source_reader_data, "get", lambda *_: "")("path") or "")
        proxy_path = str(getattr(self.proxy_reader_data, "get", lambda *_: "")("path") or "")
        if self.file_path and self.file_path == proxy_path and proxy_path != source_path:
            reader_kind = "proxy_reader"
        else:
            reader_kind = "source_reader"
        log.debug(
            "Preview dialog opening with %s path=%s size=%sx%s",
            reader_kind,
            self.file_path,
            int(reader_data.get("width", 0) or 0),
            int(reader_data.get("height", 0) or 0),
        )

        self.clip = openshot.Clip(self.file_path)
        if not is_single_image:
            self.clip.SetJson(json.dumps({
                "reader_orientation_mode": "reader",
                "reader": self.reader_data,
            }))
        self.clip.Open()

        c_info = self.clip.Reader().info
        self.width = int(getattr(c_info, "width", self.width))
        self.height = int(getattr(c_info, "height", self.height))
        fps_info = getattr(c_info, "fps", None)
        self.fps_num = int(getattr(fps_info, "num", self.fps_num))
        self.fps_den = int(getattr(fps_info, "den", self.fps_den))
        self.fps = float(self.fps_num) / float(self.fps_den or 1)
        project = getattr(get_app(), "project", None)
        if project:
            self.sample_rate = int(project.get("sample_rate"))
            self.channels = int(project.get("channels"))
            self.channel_layout = int(project.get("channel_layout"))

        base_width = max(2, int(self.width or max_size.width() or 2))
        base_height = max(2, int(self.height or max_size.height() or 2))
        if is_single_image:
            project = getattr(get_app(), "project", None)
            if project:
                base_width = max(base_width, int(project.get("width") or 0))
                base_height = max(base_height, int(project.get("height") or 0))

        aspect_ratio = openshot.Fraction(base_width, base_height)
        aspect_ratio.Reduce()
        self.videoPreview.aspect_ratio = aspect_ratio
        viewport_rect = self.videoPreview.centeredViewport(self.videoPreview.width(), self.videoPreview.height())
        max_size = QSize(max(2, int(viewport_rect.width())), max(2, int(viewport_rect.height())))

        self.r = openshot.Timeline(
            base_width,
            base_height,
            openshot.Fraction(self.fps_num, self.fps_den),
            self.sample_rate,
            self.channels,
            self.channel_layout)
        self.r.info.channel_layout = self.channel_layout
        self.r.SetMaxSize(max_size.width(), max_size.height())

        self.clip.Start(self.file.data.get("start", 0.0))
        self.clip.End(self.file.data.get("end", self.file.data.get("duration", 0.0)))

        if not self.clip.Reader().info.has_video and self.clip.Reader().info.has_audio:
            self.clip.Waveform(True)

        self.r.info.has_audio = self.clip.Reader().info.has_audio
        self.r.info.video_length = self.video_length
        self.clip.display = openshot.FRAME_DISPLAY_CLIP
        self.r.AddClip(self.clip)
        self.r.Open()

    def _reload_preview_reader(self, reader_data, max_size):
        current_frame = max(1, int(self.sliderVideo.value() or 1))
        old_preview_parent = getattr(self, "preview_parent", None)
        old_timeline = getattr(self, "r", None)
        old_clip = getattr(self, "clip", None)
        was_playing = False
        try:
            was_playing = (
                getattr(self, "preview_thread", None) is not None
                and self.preview_thread.player.Mode() == openshot.PLAYBACK_PLAY
                and self.preview_thread.player.Speed() != 0.0
            )
        except Exception:
            was_playing = False

        self.initialized = False
        if old_preview_parent:
            try:
                old_preview_parent.Stop(wait_for_thread=True)
            except Exception as exc:
                log.warning("Failed to stop previous cutting preview thread cleanly: %s", exc)

        self._build_preview_timeline(reader_data, max_size)

        self.preview_parent = PreviewParent()
        self.preview_parent.Init(self, self.r, self.videoPreview, self.video_length)
        self.preview_thread = self.preview_parent.worker
        self.initialized = True
        self.LoadTimelineAndSeekSignal.emit(current_frame)
        if was_playing:
            QTimer.singleShot(0, lambda: self.btnPlay_clicked(force="play"))

        if old_timeline:
            try:
                old_timeline.Close()
                old_timeline.ClearAllCache(True)
            except Exception:
                pass
        if old_clip:
            try:
                old_clip.Close()
            except Exception:
                pass

    def _apply_dynamic_preview_max_size(self):
        if not getattr(self, "initialized", False) or not getattr(self, "r", None):
            return

        new_size = self._target_preview_max_size()
        desired_reader_data = self._select_reader_data_for_size(new_size)
        if str(desired_reader_data.get("path") or "") != str(getattr(self, "reader_data", {}).get("path") or ""):
            self._reload_preview_reader(desired_reader_data, new_size)
            return

        previous_width = int(getattr(self.r, "preview_width", 0) or 0)
        previous_height = int(getattr(self.r, "preview_height", 0) or 0)

        if previous_width == new_size.width() and previous_height == new_size.height():
            return

        was_playing = False
        try:
            was_playing = (
                getattr(self, "preview_thread", None) is not None
                and self.preview_thread.player.Mode() == openshot.PLAYBACK_PLAY
                and self.preview_thread.player.Speed() != 0.0
            )
        except Exception:
            was_playing = False

        self.PauseSignal.emit()
        self.r.SetMaxSize(new_size.width(), new_size.height())
        self.r.ClearAllCache(True)
        self.refreshFrameSignal.emit()
        if was_playing:
            QTimer.singleShot(0, lambda: self.btnPlay_clicked(force="play"))

    def eventFilter(self, obj, event):
        if event.type() == QEvent.KeyPress and obj is self.txtName:
            # Handle ENTER key to create new clip
            if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
                if self.btnAddClip.isEnabled():
                    self.btnAddClip_clicked()
                    return True
        if event.type() == QEvent.MouseButtonPress and isinstance(obj, QSlider):
            # Handle QSlider click, jump to cursor position
            if event.button() == Qt.LeftButton:
                min_val = obj.minimum()
                max_val = obj.maximum()

                click_position = event.pos().x() if obj.orientation() == Qt.Horizontal else event.pos().y()
                slider_length = obj.width() if obj.orientation() == Qt.Horizontal else obj.height()
                new_value = min_val + ((max_val - min_val) * click_position) / slider_length

                obj.setValue(int(new_value))
                event.accept()
        return super().eventFilter(obj, event)

    def actionPlay_Triggered(self):
        # Trigger play button (This action is invoked from the preview thread, so it must exist here)
        self.btnPlay.click()

    def frame_to_timestamp(self, frame_number):
        """Return a timecode string for the given frame"""
        seconds = (frame_number - 1) / self.fps
        t = time_parts.secondsToTime(seconds, self.fps_num, self.fps_den)
        return "%s:%s:%s:%s" % (t["hour"], t["min"], t["sec"], t["frame"])

    def frame_to_compact_timestamp(self, frame_number, include_hours=False, include_minutes=False):
        """Return a compact timecode for split clip names"""
        seconds = (frame_number - 1) / self.fps
        t = time_parts.secondsToTime(seconds, self.fps_num, self.fps_den)

        hours = int(t["hour"])
        minutes = int(t["min"])
        secs = int(t["sec"])
        frames = int(t["frame"])

        if include_hours:
            return f"{hours:02d}:{minutes:02d}:{secs:02d};{frames:02d}"
        if include_minutes:
            return f"{minutes:02d}:{secs:02d};{frames:02d}"
        return f"{secs:02d};{frames:02d}"

    def movePlayhead(self, frame_number):
        """Update the playhead position"""

        # Keep slider drag native; ignore async playhead pushes while dragging.
        if self.sliderVideo.isSliderDown():
            self.lblVideoTime.setText(self.frame_to_timestamp(self.sliderVideo.value()))
            return

        # Move slider to correct frame position
        self.sliderIgnoreSignal = True
        self.sliderVideo.setValue(frame_number)
        self.sliderIgnoreSignal = False

        # Update label
        self.lblVideoTime.setText(self.frame_to_timestamp(frame_number))

    def btnPlay_clicked(self, force=None):
        log.info("btnPlay_clicked")

        if force is None and self._preview_autoplay_active:
            # Respect explicit user input (don't keep forcing startup autoplay).
            self._preview_autoplay_active = False

        if force == "pause":
            self.btnPlay.setChecked(False)
        elif force == "play":
            self.btnPlay.setChecked(True)

        if self.btnPlay.isChecked():
            log.info('play (icon to pause)')
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-pause")
            # In non-loop mode, replay from the beginning when currently at end.
            if not self.loop_playback:
                try:
                    current_pos = int(self.preview_thread.player.Position())
                except Exception:
                    current_pos = 1
                if current_pos >= int(self.video_length):
                    self.SeekSignal.emit(1)
            self.PlaySignal.emit()
        else:
            log.info('pause (icon to play)')
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-start")  # to default
            self.PauseSignal.emit()

        # Send focus back to toolbar
        self.sliderVideo.setFocus()

    def _start_preview_autoplay(self):
        if not self._preview_autoplay_active or not self.is_preview_mode:
            return
        if not getattr(self, "initialized", False):
            QTimer.singleShot(120, lambda: Cutting._start_preview_autoplay(self))
            return
        if not getattr(self, "preview_thread", None):
            QTimer.singleShot(120, lambda: Cutting._start_preview_autoplay(self))
            return
        if (
            getattr(self, "videoPreview", None) is not None
            and getattr(self.videoPreview, "delayed_resize_timer", None) is not None
            and self.videoPreview.delayed_resize_timer.isActive()
        ):
            # Let preview resize/max-size settling finish before forcing the
            # first autoplay attempt, otherwise startup can pause/clear the
            # preview cache mid-play and produce an audible glitch.
            QTimer.singleShot(120, lambda: Cutting._start_preview_autoplay(self))
            return
        if self._preview_autoplay_attempts >= 30:
            self._preview_autoplay_active = False
            return

        self._preview_autoplay_attempts += 1
        self.btnPlay_clicked(force="play")

        QTimer.singleShot(120, lambda: Cutting._start_preview_autoplay(self))

    def _preview_ready(self):
        if not self.is_preview_mode:
            return
        if getattr(self, "preview_thread", None):
            # Startup preview should show frame 1 immediately without waiting for
            # preroll/cache work before the first autoplay attempt.
            self.preview_thread.Seek(1, False)
        else:
            self.SeekSignal.emit(1)
        if self._preview_autoplay_active:
            QTimer.singleShot(0, self._start_preview_autoplay)

    def _preview_mode_changed(self, mode):
        play_mode = getattr(openshot, "PLAYBACK_PLAY", None)
        paused_mode = getattr(openshot, "PLAYBACK_PAUSED", getattr(openshot, "PLAYBACK_PAUSE", None))
        stop_mode = getattr(openshot, "PLAYBACK_STOPPED", getattr(openshot, "PLAYBACK_STOP", None))

        # Keep the play button state visually in sync with current playback mode.
        if mode == play_mode and not self.btnPlay.isChecked():
            self.btnPlay.setChecked(True)
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-pause")
        if mode == play_mode and self._preview_autoplay_active and self._preview_autoplay_attempts > 0:
            self._preview_autoplay_active = False
        elif mode in (paused_mode, stop_mode) and self.btnPlay.isChecked():
            self.btnPlay.setChecked(False)
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-start")

        if not self.is_preview_mode or not self._preview_autoplay_active:
            return

    def sliderVideo_valueChanged(self, new_frame):
        if self.preview_thread and not self.sliderIgnoreSignal:
            log.info('sliderVideo_valueChanged')
            # Pause video and update preview immediately
            self.btnPlay_clicked(force="pause")
            self.previewFrameSignal.emit(new_frame)
            # Start timer to ensure preview updates after dragging stops
            self.slider_timer.start()

    def sliderVideo_timeout(self):
        if self.preview_thread and not self.sliderIgnoreSignal:
            log.info('sliderVideo_timeout')
            self.btnPlay_clicked(force="pause")
            self.previewFrameSignal.emit(self.sliderVideo.value())

    def sliderVideo_released(self):
        if self.preview_thread and not self.sliderIgnoreSignal:
            log.info('sliderVideo_released')
            self.btnPlay_clicked(force="pause")
            self.previewFrameSignal.emit(self.sliderVideo.value())
            self.slider_timer.start()

    def btnStart_clicked(self):
        """Start of clip button was clicked"""
        _ = get_app()._tr

        # Pause video
        self.btnPlay_clicked(force="pause")

        # Get the current frame
        current_frame = self.sliderVideo.value()

        # Check if starting frame less than end frame
        if self.btnEnd.isEnabled() and current_frame >= self.end_frame:
            # Handle exception
            msg = QMessageBox()
            msg.setText(_("Please choose valid 'start' and 'end' values for your clip."))
            msg.exec_()
            return

        # remember frame #
        self.start_frame = current_frame

        # Save thumbnail image
        self.start_image = os.path.join(info.USER_PATH, 'thumbnail', '%s.png' % self.start_frame)
        self.r.GetFrame(self.start_frame).Thumbnail(self.start_image, 160, 90, '', '', '#000000', True, 'png', 85)

        # Set CSS on button
        self.btnStart.setStyleSheet('background-image: url(%s);' % self.start_image.replace('\\', '/'))

        # Enable end button
        self.btnEnd.setEnabled(True)
        self.btnClear.setEnabled(True)

        # Send focus back to toolbar
        self.sliderVideo.setFocus()

        log.info('btnStart_clicked, current frame: %s' % self.start_frame)

    def btnEnd_clicked(self):
        """End of clip button was clicked"""
        _ = get_app()._tr

        # Pause video
        self.btnPlay_clicked(force="pause")

        # Get the current frame
        current_frame = self.sliderVideo.value()

        # Check if ending frame greater than start frame
        if current_frame <= self.start_frame:
            # Handle exception
            msg = QMessageBox()
            msg.setText(_("Please choose valid 'start' and 'end' values for your clip."))
            msg.exec_()
            return

        # remember frame #
        self.end_frame = current_frame

        # Save thumbnail image
        self.end_image = os.path.join(info.USER_PATH, 'thumbnail', '%s.png' % self.end_frame)
        self.r.GetFrame(self.end_frame).Thumbnail(self.end_image, 160, 90, '', '', '#000000', True, 'png', 85)

        # Set CSS on button
        self.btnEnd.setStyleSheet('background-image: url(%s);' % self.end_image.replace('\\', '/'))

        # Enable create button
        self.btnAddClip.setEnabled(True)

        # Send focus back to toolbar
        self.sliderVideo.setFocus()

        log.info('btnEnd_clicked, current frame: %s' % self.end_frame)

    def btnClear_clicked(self):
        """Clear the current clip and reset the form"""
        log.info('btnClear_clicked')

        # Reset form
        self.clearForm()

    def clearForm(self):
        """Clear all form controls"""
        # Clear buttons
        self.start_frame = 1
        self.end_frame = 1
        self.start_image = ''
        self.end_image = ''
        self.btnStart.setStyleSheet('background-image: None;')
        self.btnEnd.setStyleSheet('background-image: None;')

        # Clear text
        self.txtName.setText('')

        # Disable buttons
        self.btnEnd.setEnabled(False)
        self.btnAddClip.setEnabled(False)
        self.btnClear.setEnabled(False)

    def btnAddClip_clicked(self):
        """Add the selected clip to the project"""
        log.info('btnAddClip_clicked')

        # Remove unneeded attributes
        if 'name' in self.file.data:
            self.file.data.pop('name')

        # Save new file
        self.file.id = None
        self.file.key = None
        self.file.type = 'insert'
        self.file.data['start'] = self.previous_start + ((self.start_frame - 1) / self.fps)
        self.file.data['end'] = self.previous_start + (self.end_frame / self.fps)
        if self.txtName.text():
            self.file.data['name'] = self.txtName.text()
        else:
            global_start_frame = round(self.previous_start * self.fps) + self.start_frame
            global_end_frame = round(self.previous_start * self.fps) + self.end_frame
            start_seconds = (global_start_frame - 1) / self.fps
            end_seconds = (global_end_frame - 1) / self.fps
            start_parts = time_parts.secondsToTime(start_seconds, self.fps_num, self.fps_den)
            end_parts = time_parts.secondsToTime(end_seconds, self.fps_num, self.fps_den)

            include_hours = int(start_parts["hour"]) > 0 or int(end_parts["hour"]) > 0
            include_minutes = (
                include_hours
                or int(start_parts["min"]) > 0
                or int(end_parts["min"]) > 0
            )
            start_timestamp = self.frame_to_compact_timestamp(
                global_start_frame, include_hours=include_hours, include_minutes=include_minutes)
            end_timestamp = self.frame_to_compact_timestamp(
                global_end_frame, include_hours=include_hours, include_minutes=include_minutes)
            base = os.path.splitext(os.path.basename(self.file_path))[0]
            self.file.data['name'] = f"{base} ({start_timestamp} to {end_timestamp})"
        self.file.save()

        # Move to next frame
        self.sliderVideo.setValue(self.end_frame + 1)

        # Reset form
        self.clearForm()

    def _finalize_preview_shutdown(self):
        if getattr(self, "videoPreview", None):
            get_app().updates.disconnect_listener(self.videoPreview)
            self.videoPreview.deleteLater()
            self.videoPreview = None

        if getattr(self, "r", None):
            try:
                self.r.Close()
                self.r.ClearAllCache()
            except Exception:
                pass
        if getattr(self, "clip", None):
            try:
                self.clip.Close()
            except Exception:
                pass

        self.preview_parent = None
        self.preview_thread = None

    def _on_preview_stopped(self):
        self._finalize_preview_shutdown()
        self._shutdown_in_progress = False
        if self._close_after_shutdown:
            self._close_after_shutdown = False
            super().reject()

    def _shutdown_preview(self, close_dialog=False):
        # Stop playback and preview worker safely (used by ESC/reject and close).
        if close_dialog:
            self._close_after_shutdown = True

        if self._shutdown_in_progress:
            return
        self._shutdown_in_progress = True

        if getattr(self, "preview_thread", None):
            try:
                self.PauseSignal.emit()
                self.StopSignal.emit()
            except Exception:
                pass

        if getattr(self, "preview_parent", None):
            background = getattr(self.preview_parent, "background", None)
            if background and background.isRunning():
                try:
                    background.finished.disconnect(self._on_preview_stopped)
                except Exception:
                    pass
                background.finished.connect(self._on_preview_stopped)
                try:
                    self.preview_parent.Stop(wait_for_thread=False)
                    return
                except Exception:
                    pass

            try:
                self.preview_parent.Stop(wait_for_thread=True)
            except Exception:
                pass

        self._on_preview_stopped()

    def reject(self):
        log.debug('reject')
        self.hide()
        self._shutdown_preview(close_dialog=True)

    def closeEvent(self, event):
        log.debug('closeEvent')
        event.ignore()
        self.reject()



"""
@file
@brief Unit tests for thumbnail generation fallback behavior.
"""

import unittest
from types import SimpleNamespace
from unittest.mock import patch

import classes.thumbnail as thumbnail


class _Frame:
    def __init__(self):
        self.calls = []

    def Thumbnail(self, *args, **kwargs):
        self.calls.append((args, kwargs))


class _Reader:
    def __init__(self, *, open_error=None, metadata=None):
        self.open_error = open_error
        self.open_calls = 0
        self.close_calls = 0
        self.decode_sizes = []
        self.max_frames = []
        self.metadata = metadata or {}

        class _Metadata(dict):
            def count(self, key):
                return 1 if key in self else 0

        self.info = SimpleNamespace(metadata=_Metadata(self.metadata))
        self._frame = _Frame()

    def SetMaxDecodeSize(self, width, height):
        self.decode_sizes.append((width, height))

    def Open(self):
        self.open_calls += 1
        if self.open_error:
            raise self.open_error

    def GetFrame(self, _):
        self.max_frames.append(_)
        return self._frame

    def Close(self):
        self.close_calls += 1


class _QueriedFile:
    def __init__(self):
        self.data = {"fps": {"num": 25, "den": 1}}

    def absolute_path(self):
        return "/tmp/dummy.webp"


class ThumbnailGenerationTests(unittest.TestCase):
    def test_generate_thumbnail_retries_when_first_reader_fails_to_open(self):
        first_reader = _Reader(open_error=RuntimeError("QtImageReader could not open image file."))
        second_reader = _Reader()
        created = []
        reader_iter = iter([first_reader, second_reader])

        def create_reader(path, inspect_reader):
            created.append((path, inspect_reader))
            return next(reader_iter)

        with patch.object(thumbnail.openshot.Clip, "CreateReader", side_effect=create_reader), \
             patch.object(thumbnail.os.path, "exists", side_effect=lambda path: True), \
             patch.object(thumbnail.shutil, "copyfile") as copyfile_patch:
            thumbnail.GenerateThumbnail("image.webp", "/tmp/thumb.png", 1, 20, 20, None, None)

        self.assertEqual(created, [("image.webp", False), ("image.webp", True)])
        self.assertEqual(first_reader.open_calls, 1)
        self.assertEqual(second_reader.open_calls, 1)
        self.assertEqual(second_reader.decode_sizes, [(60, 60)])
        self.assertEqual(second_reader.max_frames, [1])
        self.assertEqual(first_reader.close_calls, 1)
        self.assertEqual(second_reader.close_calls, 1)
        copyfile_patch.assert_not_called()

    def test_generate_thumbnail_renders_svg_placeholder_when_source_cannot_open(self):
        failing_readers = [
            _Reader(open_error=RuntimeError("missing source")),
            _Reader(open_error=RuntimeError("missing source")),
        ]
        placeholder_reader = _Reader()
        reader_iter = iter([*failing_readers, placeholder_reader])
        created = []

        def create_reader(path, inspect_reader):
            created.append((path, inspect_reader))
            return next(reader_iter)

        with patch.object(thumbnail.openshot.Clip, "CreateReader", side_effect=create_reader), \
             patch.object(thumbnail.os.path, "exists", side_effect=lambda path: True), \
             patch.object(thumbnail.shutil, "copyfile") as copyfile_patch:
            thumbnail.GenerateThumbnail("missing.webp", "/tmp/thumb.png", 7, 20, 20, None, None)

        self.assertEqual(created[0:2], [("missing.webp", False), ("missing.webp", True)])
        self.assertEqual(created[2], (thumbnail.os.path.join(thumbnail.info.IMAGES_PATH, "NotFound.svg"), False))
        self.assertEqual(placeholder_reader.max_frames, [1])
        copyfile_patch.assert_not_called()

    def test_http_thumbnail_handler_uses_project_file_icon_size(self):
        file_record = _QueriedFile()
        handler = thumbnail.httpThumbnailHandler.__new__(thumbnail.httpThumbnailHandler)
        handler.path = "/thumbnails/file-id/1/path/"
        handler.wfile = type("_Writer", (), {"write": lambda *_: None})()
        handler.send_response_only = unittest.mock.Mock()
        handler.send_header = unittest.mock.Mock()
        handler.end_headers = unittest.mock.Mock()
        handler.send_error = unittest.mock.Mock()

        with patch.object(thumbnail, "File") as file_cls, \
             patch.object(thumbnail, "GenerateThumbnail") as generate_thumbnail, \
             patch.object(thumbnail, "ThumbnailCacheIsStale", return_value=False), \
             patch.object(thumbnail.os.path, "exists", return_value=False), \
             patch.object(thumbnail.time, "sleep", return_value=None):
            file_cls.get.return_value = file_record
            handler.do_GET()

        self.assertEqual(generate_thumbnail.call_count, 1)
        call_kwargs = generate_thumbnail.call_args[0]
        self.assertEqual(call_kwargs[3], thumbnail.info.LIST_ICON_SIZE.width())
        self.assertEqual(call_kwargs[4], thumbnail.info.LIST_ICON_SIZE.height())

    def test_http_thumbnail_handler_regenerates_stale_project_file_thumbnail(self):
        file_record = _QueriedFile()
        handler = thumbnail.httpThumbnailHandler.__new__(thumbnail.httpThumbnailHandler)
        handler.path = "/thumbnails/file-id/1/path/"
        handler.wfile = type("_Writer", (), {"write": lambda *_: None})()
        handler.send_response_only = unittest.mock.Mock()
        handler.send_header = unittest.mock.Mock()
        handler.end_headers = unittest.mock.Mock()
        handler.send_error = unittest.mock.Mock()

        with patch.object(thumbnail, "File") as file_cls, \
             patch.object(thumbnail, "GenerateThumbnail") as generate_thumbnail, \
             patch.object(thumbnail, "ThumbnailCacheIsStale", return_value=True), \
             patch.object(thumbnail.os.path, "exists", return_value=True), \
             patch.object(thumbnail.time, "sleep", return_value=None):
            file_cls.get.return_value = file_record
            handler.do_GET()

        self.assertEqual(generate_thumbnail.call_count, 1)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for tutorial target handling
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import types
import sys
import unittest
from unittest.mock import MagicMock, patch

metrics = types.ModuleType("classes.metrics")
metrics.track_metric_screen = MagicMock()
with patch.dict(sys.modules, {"classes.metrics": metrics}):
    from windows.views import tutorial as tutorial_module

TutorialManager = tutorial_module.TutorialManager


class TutorialManagerTests(unittest.TestCase):
    def make_manager(self, tutorial_objects, targets):
        manager = types.SimpleNamespace(
            current_dialog=None,
            tutorial_objects=tutorial_objects,
            tutorial_enabled=True,
            tutorial_ids=[],
            position_widget=None,
            offset=None,
            win=MagicMock(),
            get_object=lambda object_id: targets.get(object_id),
            re_show_dialog=MagicMock(),
            re_position_dialog=MagicMock(),
            next_tip=MagicMock(),
            hide_tips=MagicMock(),
        )
        return manager

    def test_get_play_object_tolerates_missing_pause_action(self):
        action_play = MagicMock()
        action_play.associatedWidgets.return_value = []
        manager = types.SimpleNamespace(
            win=types.SimpleNamespace(actionPlay=action_play),
        )
        manager._get_associated_widgets = (
            lambda action: TutorialManager._get_associated_widgets(manager, action)
        )

        tutorial_object = TutorialManager.get_object(manager, "actionPlay")

        self.assertIsNone(tutorial_object)
        action_play.associatedWidgets.assert_called_once_with()

    def test_process_ignores_unavailable_target(self):
        manager = self.make_manager(
            [{"id": "3", "object_id": "actionPlay"}],
            {},
        )

        TutorialManager.process(manager)

        self.assertIsNone(manager.current_dialog)
        manager.re_show_dialog.assert_not_called()

    @patch.object(tutorial_module.QTimer, "singleShot")
    @patch.object(tutorial_module, "TutorialDialog")
    def test_process_skips_unavailable_target_and_shows_next(
            self, mock_dialog_class, mock_single_shot):
        visible_target = MagicMock()
        visible_target.visibleRegion.return_value.isEmpty.return_value = False
        tutorials = [
            {"id": "3", "object_id": "actionPlay"},
            {
                "id": "4",
                "object_id": "propertyTableView",
                "x": 0,
                "y": 0,
                "text": "Properties",
                "arrow": True,
            },
        ]
        manager = self.make_manager(
            tutorials,
            {"propertyTableView": visible_target},
        )
        dialog = mock_dialog_class.return_value

        TutorialManager.process(manager)

        self.assertIs(manager.position_widget, visible_target)
        self.assertIs(manager.current_dialog, dialog)
        manager.re_show_dialog.assert_called_once()
        mock_single_shot.assert_called_once_with(0, manager.re_position_dialog)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief This file is used to generate a Final Cut Pro export
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import pathlib
import shutil
import subprocess
from operator import itemgetter
from uuid import uuid1
from xml.dom import minidom

import openshot
from qt_api import QFileDialog

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.path_utils import absolute_media_path, relative_export_path, normalize_path
from classes.query import Clip, Track, File

INTERPOLATION_EXPORT_MAP = {
    openshot.LINEAR: "linear",
    openshot.BEZIER: "bezier",
    openshot.CONSTANT: "constant"
}


def _set_text(node, value):
    """Safely set the text content of a DOM node."""
    if not node:
        return
    if node.firstChild:
        node.firstChild.nodeValue = str(value)
    else:
        node.appendChild(node.ownerDocument.createTextNode(str(value)))


def _is_ntsc_rate(fps_num, fps_den):
    return fps_den == 1001 and fps_num in (24000, 30000, 60000)


def _seconds_to_frames(value, fps_num, fps_den):
    if value is None:
        value = 0.0
    return int(round(float(value) * fps_num / fps_den))


def _format_timebase(fps_num, fps_den):
    fps_value = float(fps_num) / float(fps_den)
    text = ("{0:.3f}".format(fps_value)).rstrip('0').rstrip('.')
    return text or "0"


def _format_ratio(num, den):
    """Format a numeric ratio as a compact string."""
    try:
        num = float(num)
        den = float(den)
        if den == 0:
            raise ZeroDivisionError()
        value = num / den
        text = ("{0:.10f}".format(value)).rstrip('0').rstrip('.')
        return text or "1"
    except Exception:
        return "1"


def _displayformat(ntsc_value):
    """Choose drop-frame vs non-drop timecode string."""
    ntsc_upper = (ntsc_value or "").upper()
    return "DF" if ntsc_upper == "TRUE" else "NDF"


def _timecode_string(ntsc_value):
    """Return a default timecode string respecting drop/non-drop separators."""
    separator = ";" if (ntsc_value or "").upper() == "TRUE" else ":"
    return f"00{separator}00{separator}00{separator}00"


def _file_url(path_string):
    """Return a file:// URL for an absolute path if possible."""
    if not path_string:
        return ""
    try:
        return pathlib.Path(path_string).absolute().as_uri()
    except Exception:
        return normalize_path(path_string)


def _unique_file_id(path_string, existing_id, id_map):
    """Ensure a unique file id per absolute path."""
    normalized = os.path.abspath(path_string or "")
    if normalized in id_map:
        return id_map[normalized]
    candidate = existing_id or "file-%s" % uuid1()
    if candidate in id_map.values():
        candidate = "file-%s" % uuid1()
    id_map[normalized] = candidate
    return candidate


def _clip_file_info(clip):
    """Return (File object, merged metadata dict, absolute file path)."""
    file_obj = None
    file_data = {}
    file_id = clip.data.get("file_id")
    if file_id:
        file_obj = File.get(id=file_id)
        if file_obj and file_obj.data:
            file_data = dict(file_obj.data)

    reader_data = clip.data.get("reader", {}) or {}
    merged_data = dict(reader_data)
    merged_data.update(file_data)

    file_path = ""
    if file_obj:
        file_path = file_obj.absolute_path()
    if not file_path:
        file_path = absolute_media_path(reader_data.get("path"))
    if not file_path:
        file_path = reader_data.get("path", "")

    return file_obj, merged_data, file_path


def _apply_rate_settings(rate_nodes, timebase_value, ntsc_value):
    for rate_node in rate_nodes:
        timebase_nodes = rate_node.getElementsByTagName("timebase")
        if timebase_nodes:
            _set_text(timebase_nodes[0], timebase_value)
        ntsc_nodes = rate_node.getElementsByTagName("ntsc")
        if ntsc_nodes:
            _set_text(ntsc_nodes[0], ntsc_value)


def _apply_timecode_settings(timecode_nodes, timebase_value, ntsc_value):
    for timecode_node in timecode_nodes:
        _apply_rate_settings(timecode_node.getElementsByTagName("rate"), timebase_value, ntsc_value)
        displayformat_nodes = timecode_node.getElementsByTagName("displayformat")
        if displayformat_nodes:
            _set_text(displayformat_nodes[0], _displayformat(ntsc_value))
        string_nodes = timecode_node.getElementsByTagName("string")
        if string_nodes:
            _set_text(string_nodes[0], _timecode_string(ntsc_value))


def _validate_export(xml_path):
    """Validate the exported XML against the bundled Final Cut Pro DTDs."""
    xmllint_path = shutil.which("xmllint")
    if not xmllint_path:
        log.info("Skipping FCP XML validation; xmllint not available")
        return

    for dtd_file in ("fcp-xml-v4.dtd", "fcp-xml-v5.dtd"):
        dtd_path = os.path.join(info.RESOURCES_PATH, dtd_file)
        if not os.path.exists(dtd_path):
            continue

        result = subprocess.run(
            [xmllint_path, "--noout", "--dtdvalid", dtd_path, xml_path],
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            log.info("Validated Final Cut Pro XML against %s", dtd_file)
        else:
            error_text = (result.stderr or result.stdout or "").strip()
            log.warning("Final Cut Pro XML did not validate against %s: %s", dtd_file, error_text)


def _append_link(link_parent, mediatype_value, track_index, clip_index, group_index, ref_id):
    """Append a <link> node for A/V grouping compatibility."""
    if not link_parent:
        return
    link_node = link_parent.ownerDocument.createElement("link")
    link_parent.appendChild(link_node)
    for tag, value in (
        ("linkclipref", ref_id),
        ("mediatype", mediatype_value),
        ("trackindex", track_index),
        ("clipindex", clip_index),
        ("groupindex", group_index),
    ):
        child = link_parent.ownerDocument.createElement(tag)
        child.appendChild(link_parent.ownerDocument.createTextNode(str(value)))
        link_node.appendChild(child)


def _normalized_to_center_pixels(x_norm, y_norm, frame_width, frame_height):
    """Map normalized OpenShot coords (-1..1, origin center) to pixel center values."""
    try:
        w = float(frame_width)
        h = float(frame_height)
    except (TypeError, ValueError):
        return x_norm, y_norm
    if w <= 0 or h <= 0:
        return x_norm, y_norm
    return (w / 2.0) + (x_norm * w / 2.0), (h / 2.0) + (y_norm * h / 2.0)


def _export_interp_name(value):
    """Return an interpolation name for export from int/str values."""
    if isinstance(value, str):
        val = value.strip().lower()
        if val in ("linear", "0"):
            return "linear"
        if val in ("bezier", "ease", "easein", "easeout", "1"):
            return "bezier"
        if val in ("hold", "constant", "2"):
            return "constant"
    try:
        return INTERPOLATION_EXPORT_MAP.get(int(value), "linear")
    except Exception:
        return "linear"


def _scale_mode_size(src_w, src_h, frame_w, frame_h, scale_mode):
    """Return base scaled dimensions after applying scale mode (before per-axis scale)."""
    try:
        sw = float(src_w)
        sh = float(src_h)
        fw = float(frame_w)
        fh = float(frame_h)
    except (TypeError, ValueError):
        return src_w, src_h
    if sw <= 0 or sh <= 0 or fw <= 0 or fh <= 0:
        return src_w, src_h
    if scale_mode == openshot.SCALE_STRETCH:
        return fw, fh
    if scale_mode == openshot.SCALE_CROP:
        factor = max(fw / sw, fh / sh)
        return sw * factor, sh * factor
    if scale_mode == openshot.SCALE_FIT:
        factor = min(fw / sw, fh / sh)
        return sw * factor, sh * factor
    # SCALE_NONE or unknown
    return sw, sh


def _gravity_offset(gravity, frame_w, frame_h, scaled_w, scaled_h):
    """Top-left origin based on gravity inside the frame."""
    try:
        frame_w = float(frame_w)
        frame_h = float(frame_h)
        scaled_w = float(scaled_w)
        scaled_h = float(scaled_h)
    except (TypeError, ValueError):
        return 0.0, 0.0
    x = 0.0
    y = 0.0
    if gravity == openshot.GRAVITY_TOP:
        x = (frame_w - scaled_w) / 2.0
    elif gravity == openshot.GRAVITY_TOP_RIGHT:
        x = frame_w - scaled_w
    elif gravity == openshot.GRAVITY_LEFT:
        y = (frame_h - scaled_h) / 2.0
    elif gravity == openshot.GRAVITY_CENTER:
        x = (frame_w - scaled_w) / 2.0
        y = (frame_h - scaled_h) / 2.0
    elif gravity == openshot.GRAVITY_RIGHT:
        x = frame_w - scaled_w
        y = (frame_h - scaled_h) / 2.0
    elif gravity == openshot.GRAVITY_BOTTOM_LEFT:
        y = frame_h - scaled_h
    elif gravity == openshot.GRAVITY_BOTTOM:
        x = (frame_w - scaled_w) / 2.0
        y = frame_h - scaled_h
    elif gravity == openshot.GRAVITY_BOTTOM_RIGHT:
        x = frame_w - scaled_w
        y = frame_h - scaled_h
    return x, y


def _merge_uniform_scale(scale_x_points, scale_y_points):
    """Combine X/Y scale into a single uniform curve (average when both exist)."""
    merged = {}

    def _add(points):
        for point in points:
            t = point.get("co", {}).get("X", 1)
            v = point.get("co", {}).get("Y", 1.0)
            interp = point.get("interpolation", openshot.LINEAR)
            if t in merged:
                merged[t]["val"] = (merged[t]["val"] + v) / 2.0
                if merged[t].get("interp") is None:
                    merged[t]["interp"] = interp
            else:
                merged[t] = {"val": v, "interp": interp}

    _add(scale_x_points or [])
    _add(scale_y_points or [])

    merged_points = []
    for t in sorted(merged.keys()):
        merged_points.append(
            {
                "co": {"X": t, "Y": merged[t]["val"]},
                "interpolation": merged[t].get("interp", openshot.LINEAR),
            }
        )
    return merged_points


def _find_effect_node(node, effect_name):
    """Find the first effect node with a matching name."""
    for effectNode in node.getElementsByTagName("effect"):
        effectName = effectNode.getElementsByTagName("name")[0].childNodes[0].nodeValue
        if effectName == effect_name:
            return effectNode
    return None


def _find_parameter_node(effect_node, parameter_id):
    """Find the first parameter node with a matching id."""
    for param in effect_node.getElementsByTagName("parameter"):
        paramIdNode = param.getElementsByTagName("parameterid")
        if paramIdNode and paramIdNode[0].childNodes and paramIdNode[0].childNodes[0].nodeValue == parameter_id:
            return param
    return None


def createEffect(xmldoc, name, node, points, scale, max_frames=None, param_name=None):
    """Create the XML filter with keyframes"""
    effectNode = _find_effect_node(node, name)
    if not effectNode:
        return
    if param_name:
        parameterNode = _find_parameter_node(effectNode, param_name)
    else:
        parameterNode = effectNode.getElementsByTagName("parameter")[0]

    if not parameterNode:
        return

    # Loop through Points (remove duplicates)
    keyframes = {}
    for point in points:
        keyframeTime = point.get('co', {}).get('X', 1)
        keyframeValue = point.get('co', {}).get('Y', 1) * scale
        interpolation = point.get('interpolation', openshot.LINEAR)
        if max_frames is not None:
            keyframeTime = max(0, min(keyframeTime, max_frames))
        keyframes[keyframeTime] = (keyframeValue, interpolation)

    # Loop through Points
    for keyframeTime in sorted(keyframes.keys()):
        keyframeValue, interpolation = keyframes.get(keyframeTime)

        # Create keyframe element for each point
        keyframeNode = xmldoc.createElement("keyframe")
        parameterNode.appendChild(keyframeNode)
        whenNode = xmldoc.createElement("when")
        whenNode.appendChild(xmldoc.createTextNode(str(int(round(keyframeTime)))))
        keyframeNode.appendChild(whenNode)
        valueNode = xmldoc.createElement("value")
        valueNode.appendChild(xmldoc.createTextNode(str(keyframeValue)))
        keyframeNode.appendChild(valueNode)
        interpNode = xmldoc.createElement("interpolation")
        interpName = xmldoc.createElement("name")
        interpName.appendChild(xmldoc.createTextNode(_export_interp_name(interpolation)))
        interpNode.appendChild(interpName)
        keyframeNode.appendChild(interpNode)


def createCenterEffect(
    xmldoc,
    node,
    x_points,
    y_points,
    frame_width,
    frame_height,
    src_width,
    src_height,
    scale_mode,
    gravity,
    scale_x_points,
    scale_y_points,
    max_frames=None,
):
    """Create the Basic Motion center keyframes using project/frame units."""
    effectNode = _find_effect_node(node, "Basic Motion")
    if not effectNode:
        return

    parameterNode = _find_parameter_node(effectNode, "center")
    if not parameterNode:
        return

    keyframes = {}

    def _add_points(points, axis_key):
        for point in points:
            keyframeTime = point.get("co", {}).get("X", 1)
            interpolation = point.get("interpolation", openshot.LINEAR)
            if max_frames is not None:
                keyframeTime = max(0, min(keyframeTime, max_frames))
            entry = keyframes.setdefault(keyframeTime, {})
            entry[axis_key] = point.get("co", {}).get("Y", 0.0)
            entry["interp_%s" % axis_key] = interpolation

    _add_points(x_points, "x")
    _add_points(y_points, "y")

    def _value_at_time(points, t, fallback=1.0):
        if not points:
            return fallback
        sorted_points = sorted(points, key=lambda p: p.get("co", {}).get("X", 0))
        last_val = sorted_points[0].get("co", {}).get("Y", fallback)
        for p in sorted_points:
            pt_time = p.get("co", {}).get("X", 0)
            if max_frames is not None:
                pt_time = max(0, min(pt_time, max_frames))
            if pt_time <= t:
                last_val = p.get("co", {}).get("Y", fallback)
            else:
                break
        return last_val

    last_x = 0.0
    last_y = 0.0

    for keyframeTime in sorted(keyframes.keys()):
        entry = keyframes[keyframeTime]
        x_val = entry.get("x", last_x)
        y_val = entry.get("y", last_y)
        last_x = x_val
        last_y = y_val
        if "interp_x" in entry:
            interpolation = entry.get("interp_x")
        elif "interp_y" in entry:
            interpolation = entry.get("interp_y")
        else:
            interpolation = openshot.LINEAR

        base_w, base_h = _scale_mode_size(src_width, src_height, frame_width, frame_height, scale_mode)
        sx = _value_at_time(scale_x_points, keyframeTime, 1.0)
        sy = _value_at_time(scale_y_points, keyframeTime, 1.0)
        scaled_w = base_w * sx
        scaled_h = base_h * sy
        origin_x, origin_y = _gravity_offset(gravity, frame_width, frame_height, scaled_w, scaled_h)
        center_x, center_y = origin_x + (scaled_w / 2.0), origin_y + (scaled_h / 2.0)
        center_x += frame_width * x_val
        center_y += frame_height * y_val

        keyframeNode = xmldoc.createElement("keyframe")
        parameterNode.appendChild(keyframeNode)
        whenNode = xmldoc.createElement("when")
        whenNode.appendChild(xmldoc.createTextNode(str(int(round(keyframeTime)))))
        keyframeNode.appendChild(whenNode)

        valueNode = xmldoc.createElement("value")
        horizNode = xmldoc.createElement("horiz")
        horizNode.appendChild(xmldoc.createTextNode(str(center_x)))
        valueNode.appendChild(horizNode)
        vertNode = xmldoc.createElement("vert")
        vertNode.appendChild(xmldoc.createTextNode(str(center_y)))
        valueNode.appendChild(vertNode)
        keyframeNode.appendChild(valueNode)

        interpNode = xmldoc.createElement("interpolation")
        interpName = xmldoc.createElement("name")
        interpName.appendChild(xmldoc.createTextNode(_export_interp_name(interpolation)))
        interpNode.appendChild(interpName)
        keyframeNode.appendChild(interpNode)


def export_xml():
    """Export final cut pro XML file"""
    app = get_app()
    _ = app._tr

    # Get FPS info
    fps_num = get_app().project.get("fps").get("num", 24)
    fps_den = get_app().project.get("fps").get("den", 1)
    timebase_value = _format_timebase(fps_num, fps_den)
    ntsc_value = "TRUE" if _is_ntsc_rate(fps_num, fps_den) else "FALSE"
    project_data = getattr(app.project, "_data", {}) or {}
    project_interlaced = False
    try:
        project_interlaced = bool(project_data.get("interlaced_frame"))
    except Exception:
        project_interlaced = False
    project_pixel_ratio = project_data.get("pixel_ratio") or {}
    project_pixel_ratio_value = _format_ratio(project_pixel_ratio.get("num", 1), project_pixel_ratio.get("den", 1))
    try:
        project_pixel_ratio_float = float(project_pixel_ratio.get("num", 1)) / float(project_pixel_ratio.get("den", 1))
    except Exception:
        project_pixel_ratio_float = 1.0
    project_anamorphic_value = "TRUE" if abs(project_pixel_ratio_float - 1.0) > 0.0005 else "FALSE"
    project_field_dominance = "lower" if project_interlaced else "none"

    # Get path
    recommended_path = get_app().project.current_filepath or ""
    if not recommended_path:
        recommended_path = os.path.join(info.HOME_PATH, "%s.xml" % _("Untitled Project"))
    else:
        recommended_path = recommended_path.replace(".osp", ".xml")
    file_path = QFileDialog.getSaveFileName(app.window, _("Export XML..."), recommended_path,
                                            _("Final Cut Pro (*.xml)"))[0]
    if not file_path:
        # User canceled dialog
        return

    # Append .xml if needed
    if not file_path.endswith(".xml"):
        file_path = "%s.xml" % file_path

    export_folder = os.path.dirname(os.path.abspath(file_path))

    # Get filename with no path
    file_name = os.path.basename(file_path)

    # Determine max frame (based on clips)
    duration = 0.0
    all_clips = Clip.filter()
    for clip in all_clips:
        clip_last_frame = (clip.data.get("position") or 0.0) + ((clip.data.get("end") or 0.0) - (clip.data.get("start") or 0.0))
        if clip_last_frame > duration:
            # Set max length of timeline
            duration = clip_last_frame
    duration_frames = _seconds_to_frames(duration, fps_num, fps_den)

    # XML template path
    xmldoc = minidom.parse(os.path.join(info.RESOURCES_PATH, 'export-project-template.xml'))

    # Set Project Details
    _set_text(xmldoc.getElementsByTagName("name")[0], file_name)
    _set_text(xmldoc.getElementsByTagName("uuid")[0], str(uuid1()))
    _set_text(xmldoc.getElementsByTagName("duration")[0], duration_frames)
    _set_text(xmldoc.getElementsByTagName("width")[0], app.project.get("width"))
    _set_text(xmldoc.getElementsByTagName("height")[0], app.project.get("height"))
    for par_node in xmldoc.getElementsByTagName("pixelaspectratio"):
        _set_text(par_node, project_pixel_ratio_value)
    for anamorphic_node in xmldoc.getElementsByTagName("anamorphic"):
        _set_text(anamorphic_node, project_anamorphic_value)
    _set_text(xmldoc.getElementsByTagName("samplerate")[0], app.project.get("sample_rate"))
    for field_node in xmldoc.getElementsByTagName("fielddominance"):
        _set_text(field_node, project_field_dominance)
    _apply_timecode_settings(xmldoc.getElementsByTagName("timecode"), timebase_value, ntsc_value)
    sequence_node = xmldoc.getElementsByTagName("sequence")[0]
    if app.project.get("id"):
        sequence_node.setAttribute("id", app.project.get("id"))
    for childNode in xmldoc.getElementsByTagName("timebase"):
        _set_text(childNode, timebase_value)
    for ntsc_node in xmldoc.getElementsByTagName("ntsc"):
        _set_text(ntsc_node, ntsc_value)

    # Get parent nodes
    parentAudioNode = xmldoc.getElementsByTagName("audio")[0]
    parentVideoNode = xmldoc.getElementsByTagName("video")[0]
    num_output_channels = parentAudioNode.getElementsByTagName("channelcount")
    if num_output_channels:
        project_channels = app.project.get("channels")
        if project_channels is None:
            project_channels = 2
        _set_text(num_output_channels[0], project_channels)

    # Loop through tracks
    all_tracks = get_app().project.get("layers")
    video_track_index = 1
    audio_track_index = 1
    clip_link_info = {}
    file_id_map = {}
    for track in sorted(all_tracks, key=itemgetter('number')):
        existing_track = Track.get(number=track.get("number"))
        if not existing_track:
            # Log error and fail silently, and continue
            log.error('No track object found with number: %s' % track.get("number"))
            continue

        # Track details
        track_locked = track.get("lock", False)
        clips_on_track = sorted(Clip.filter(layer=track.get("number")), key=lambda c: c.data.get('position', 0.0))
        if not clips_on_track:
            continue

        has_video = any(c.data.get("reader", {}).get("has_video") for c in clips_on_track)
        has_audio = any(c.data.get("reader", {}).get("has_audio") for c in clips_on_track)

        videoTrackNode = None
        audioTrackNode = None
        audioTrackNumber = None
        videoTrackNumber = None

        if has_video:
            trackTemplateDoc = minidom.parse(os.path.join(info.RESOURCES_PATH, 'export-track-video-template.xml'))
            videoTrackNode = trackTemplateDoc.getElementsByTagName('track')[0]
            parentVideoNode.appendChild(videoTrackNode)
            videoTrackNumber = video_track_index
            video_track_index += 1
            if track_locked:
                _set_text(videoTrackNode.getElementsByTagName("locked")[0], "TRUE")

        if has_audio:
            trackTemplateDoc = minidom.parse(os.path.join(info.RESOURCES_PATH, 'export-track-audio-template.xml'))
            audioTrackNode = trackTemplateDoc.getElementsByTagName('track')[0]
            parentAudioNode.appendChild(audioTrackNode)
            audioTrackNumber = audio_track_index
            audio_track_index += 1
            output_nodes = audioTrackNode.getElementsByTagName("outputchannelindex")
            if output_nodes:
                _set_text(output_nodes[0], audioTrackNumber)
            if track_locked:
                _set_text(audioTrackNode.getElementsByTagName("locked")[0], "TRUE")

        video_clip_index = 1
        audio_clip_index = 1

        # Loop through clips on this track
        for clip in clips_on_track:
            clip_reader = clip.data.get("reader", {}) or {}
            clip_duration_frames = _seconds_to_frames((clip.data.get('end') or 0.0) - (clip.data.get('start') or 0.0), fps_num, fps_den)
            if clip_duration_frames <= 0:
                continue

            clip_in_frames = _seconds_to_frames(clip.data.get('start') or 0.0, fps_num, fps_den)
            clip_out_frames = clip_in_frames + clip_duration_frames
            timeline_in_frames = _seconds_to_frames(clip.data.get('position') or 0.0, fps_num, fps_den)
            timeline_out_frames = timeline_in_frames + clip_duration_frames

            file_obj, merged_data, abs_media_path = _clip_file_info(clip)
            display_name = merged_data.get("name") or os.path.basename(merged_data.get("path", "") or clip.data.get('title') or "")
            clip_title = clip.data.get('title') or display_name or os.path.basename(abs_media_path or "") or "Clip"
            sample_rate = merged_data.get("sample_rate") or merged_data.get("audio_sample_rate") or app.project.get("sample_rate") or 48000
            channel_count = merged_data.get("channels") or merged_data.get("channel_count") or 2
            merged_pixel_ratio = merged_data.get("pixel_ratio") or project_pixel_ratio
            clip_pixel_ratio_value = _format_ratio(merged_pixel_ratio.get("num", 1), merged_pixel_ratio.get("den", 1))
            try:
                clip_pixel_ratio_float = float(merged_pixel_ratio.get("num", 1)) / float(merged_pixel_ratio.get("den", 1))
            except Exception:
                clip_pixel_ratio_float = 1.0
            clip_anamorphic_value = "TRUE" if abs(clip_pixel_ratio_float - 1.0) > 0.0005 else "FALSE"
            clip_field_dominance = project_field_dominance
            if merged_data.get("interlaced_frame"):
                clip_field_dominance = "lower"
            try:
                sample_rate = int(sample_rate)
            except (TypeError, ValueError):
                sample_rate = 48000
            try:
                channel_count = int(channel_count)
            except (TypeError, ValueError):
                channel_count = 2
            relative_media_path = relative_export_path(abs_media_path, export_folder)
            file_id_value = _unique_file_id(abs_media_path or merged_data.get("path"), clip.data.get('file_id'), file_id_map)

            # Use the single authoritative duration from the File object (if available)
            file_duration_frames = None
            if file_obj and isinstance(file_obj.data.get("duration"), (int, float)):
                file_duration_frames = _seconds_to_frames(file_obj.data.get("duration"), fps_num, fps_den)

            if file_duration_frames is not None:
                clip_in_frames = min(clip_in_frames, file_duration_frames)
                max_available = max(0, file_duration_frames - clip_in_frames)
                clip_duration_frames = min(clip_duration_frames, max_available)
                clip_out_frames = clip_in_frames + clip_duration_frames
                timeline_out_frames = timeline_in_frames + clip_duration_frames
            if clip_duration_frames <= 0:
                continue

            if clip_reader.get("has_video") and videoTrackNode:
                clipTemplateDoc = minidom.parse(os.path.join(info.RESOURCES_PATH, 'export-clip-video-template.xml'))
                clipNode = clipTemplateDoc.getElementsByTagName('clipitem')[0]
                videoTrackNode.appendChild(clipNode)

                clipNode.setAttribute('id', clip.data.get('id'))
                clip_file_node = clipNode.getElementsByTagName("file")[0]
                clip_file_node.setAttribute('id', file_id_value)
                clip_names = clipNode.getElementsByTagName("name")
                if clip_names:
                    _set_text(clip_names[0], clip_title)
                if len(clip_names) > 1:
                    _set_text(clip_names[1], clip_title)
                _set_text(clipNode.getElementsByTagName("start")[0], timeline_in_frames)
                _set_text(clipNode.getElementsByTagName("end")[0], timeline_out_frames)
                _set_text(clipNode.getElementsByTagName("in")[0], clip_in_frames)
                _set_text(clipNode.getElementsByTagName("out")[0], clip_out_frames)
                _set_text(clipNode.getElementsByTagName("duration")[0], clip_duration_frames)
                for par_node in clipNode.getElementsByTagName("pixelaspectratio"):
                    _set_text(par_node, clip_pixel_ratio_value)
                for anamorphic_node in clipNode.getElementsByTagName("anamorphic"):
                    _set_text(anamorphic_node, clip_anamorphic_value)
                _apply_rate_settings(clipNode.getElementsByTagName("rate"), timebase_value, ntsc_value)

                file_path_nodes = clip_file_node.getElementsByTagName("pathurl")
                if file_path_nodes:
                    preferred_path = abs_media_path or relative_media_path or clip_title
                    _set_text(file_path_nodes[0], _file_url(preferred_path))
                file_name_nodes = clip_file_node.getElementsByTagName("name")
                if file_name_nodes:
                    _set_text(file_name_nodes[0], display_name)
                file_duration_nodes = clip_file_node.getElementsByTagName("duration")
                if file_duration_nodes:
                    if file_duration_frames is not None:
                        _set_text(file_duration_nodes[0], file_duration_frames)
                    else:
                        _set_text(file_duration_nodes[0], clip_duration_frames)
                width_nodes = clip_file_node.getElementsByTagName("width")
                if width_nodes:
                    _set_text(width_nodes[0], merged_data.get("width") or app.project.get("width"))
                height_nodes = clip_file_node.getElementsByTagName("height")
                if height_nodes:
                    _set_text(height_nodes[0], merged_data.get("height") or app.project.get("height"))
                field_dom_nodes = clip_file_node.getElementsByTagName("fielddominance")
                for field_node in field_dom_nodes:
                    _set_text(field_node, clip_field_dominance)
                for par_node in clip_file_node.getElementsByTagName("pixelaspectratio"):
                    _set_text(par_node, clip_pixel_ratio_value)
                for anamorphic_node in clip_file_node.getElementsByTagName("anamorphic"):
                    _set_text(anamorphic_node, clip_anamorphic_value)
                _apply_timecode_settings(clip_file_node.getElementsByTagName("timecode"), timebase_value, ntsc_value)
                samplerate_nodes = clip_file_node.getElementsByTagName("samplerate")
                if samplerate_nodes:
                    _set_text(samplerate_nodes[0], sample_rate)
                channel_nodes = clip_file_node.getElementsByTagName("channelcount")
                if channel_nodes:
                    _set_text(channel_nodes[0], channel_count)
                media_nodes = clip_file_node.getElementsByTagName("media")
                if media_nodes:
                    video_nodes = media_nodes[0].getElementsByTagName("video")
                    audio_nodes = media_nodes[0].getElementsByTagName("audio")
                    if not clip_reader.get("has_audio"):
                        for n in list(audio_nodes):
                            media_nodes[0].removeChild(n)
                    if not clip_reader.get("has_video"):
                        for n in list(video_nodes):
                            media_nodes[0].removeChild(n)

                if clip_reader.get("has_video"):
                    frame_width = app.project.get("width") or merged_data.get("width")
                    frame_height = app.project.get("height") or merged_data.get("height")
                    src_width = merged_data.get("width") or frame_width
                    src_height = merged_data.get("height") or frame_height
                    scale_mode = clip.data.get("scale", openshot.SCALE_FIT)
                    gravity = clip.data.get("gravity", openshot.GRAVITY_CENTER)
                    createEffect(
                        xmldoc,
                        "Opacity",
                        clipNode,
                        clip.data.get("alpha", {}).get("Points", []),
                        100.0,
                        clip_duration_frames,
                        "opacity",
                    )
                    createCenterEffect(
                        xmldoc,
                        clipNode,
                        clip.data.get("location_x", {}).get("Points", []),
                        clip.data.get("location_y", {}).get("Points", []),
                        frame_width,
                        frame_height,
                        src_width,
                        src_height,
                        scale_mode,
                        gravity,
                        clip.data.get("scale_x", {}).get("Points", []),
                        clip.data.get("scale_y", {}).get("Points", []),
                        clip_duration_frames,
                    )
                    scale_points = _merge_uniform_scale(
                        clip.data.get("scale_x", {}).get("Points", []),
                        clip.data.get("scale_y", {}).get("Points", []),
                    )
                    createEffect(
                        xmldoc,
                        "Basic Motion",
                        clipNode,
                        scale_points,
                        100.0,
                        clip_duration_frames,
                        "scale",
                    )
                    createEffect(
                        xmldoc,
                        "Basic Motion",
                        clipNode,
                        clip.data.get("rotation", {}).get("Points", []),
                        1.0,
                        clip_duration_frames,
                        "rotation",
                    )
                logging_nodes = clipNode.getElementsByTagName("good")
                if logging_nodes:
                    _set_text(logging_nodes[0], "FALSE")
                base_id = clip.data.get('id')
                if base_id:
                    clip_link_info.setdefault(base_id, {})
                    clip_link_info[base_id]["video"] = {
                        "node": clipNode,
                        "trackindex": videoTrackNumber or 1,
                        "clipindex": video_clip_index,
                        "id": clip.data.get('id')
                    }
                    video_clip_index += 1
                    link_parts = clip_link_info[base_id]
                    if "audio" in link_parts:
                        audio_info = link_parts["audio"]
                        _append_link(clipNode, "video", videoTrackNumber or 1, link_parts["video"]["clipindex"], 1, clip.data.get('id'))
                        _append_link(clipNode, "audio", audio_info["trackindex"], audio_info["clipindex"], 1, audio_info["id"])
                        _append_link(audio_info["node"], "audio", audio_info["trackindex"], audio_info["clipindex"], 1, audio_info["id"])
                        _append_link(audio_info["node"], "video", videoTrackNumber or 1, link_parts["video"]["clipindex"], 1, clip.data.get('id'))

            if clip_reader.get("has_audio") and audioTrackNode:
                clipTemplateDoc = minidom.parse(os.path.join(info.RESOURCES_PATH, 'export-clip-audio-template.xml'))
                clipAudioNode = clipTemplateDoc.getElementsByTagName('clipitem')[0]
                audioTrackNode.appendChild(clipAudioNode)

                clipAudioNode.setAttribute('id', "%s-audio" % clip.data.get('id'))

                audio_file_node = clipAudioNode.getElementsByTagName("file")[0]
                audio_file_node.setAttribute('id', file_id_value)
                audio_name_nodes = clipAudioNode.getElementsByTagName("name")
                if audio_name_nodes:
                    _set_text(audio_name_nodes[0], clip_title)
                if len(audio_name_nodes) > 1:
                    _set_text(audio_name_nodes[1], clip_title)
                _set_text(clipAudioNode.getElementsByTagName("start")[0], timeline_in_frames)
                _set_text(clipAudioNode.getElementsByTagName("end")[0], timeline_out_frames)
                _set_text(clipAudioNode.getElementsByTagName("in")[0], clip_in_frames)
                _set_text(clipAudioNode.getElementsByTagName("out")[0], clip_out_frames)
                _set_text(clipAudioNode.getElementsByTagName("duration")[0], clip_duration_frames)
                _apply_rate_settings(clipAudioNode.getElementsByTagName("rate"), timebase_value, ntsc_value)

                audio_path_nodes = audio_file_node.getElementsByTagName("pathurl")
                if audio_path_nodes:
                    preferred_audio_path = abs_media_path or relative_media_path or clip_title
                    _set_text(audio_path_nodes[0], _file_url(preferred_audio_path))
                audio_file_names = audio_file_node.getElementsByTagName("name")
                if audio_file_names:
                    _set_text(audio_file_names[0], display_name)
                audio_file_duration_nodes = audio_file_node.getElementsByTagName("duration")
                if audio_file_duration_nodes:
                    _set_text(
                        audio_file_duration_nodes[0],
                        file_duration_frames if file_duration_frames is not None else clip_duration_frames
                    )
                samplerate_nodes = audio_file_node.getElementsByTagName("samplerate")
                if samplerate_nodes:
                    _set_text(samplerate_nodes[0], sample_rate)
                channel_nodes = audio_file_node.getElementsByTagName("channelcount")
                if channel_nodes:
                    _set_text(channel_nodes[0], channel_count)
                _apply_timecode_settings(audio_file_node.getElementsByTagName("timecode"), timebase_value, ntsc_value)

                sourcetrack_nodes = clipAudioNode.getElementsByTagName("sourcetrack")
                if sourcetrack_nodes and audioTrackNumber is not None:
                    track_index_nodes = sourcetrack_nodes[0].getElementsByTagName("trackindex")
                    if track_index_nodes:
                        _set_text(track_index_nodes[0], audioTrackNumber)

                track_index_nodes = clipAudioNode.getElementsByTagName("trackindex")
                if track_index_nodes and audioTrackNumber is not None:
                    _set_text(track_index_nodes[0], audioTrackNumber)

                createEffect(xmldoc, "Audio Levels", clipAudioNode, clip.data.get('volume', {}).get('Points', []), 1.0, clip_duration_frames)
                logging_nodes = clipAudioNode.getElementsByTagName("good")
                if logging_nodes:
                    _set_text(logging_nodes[0], "FALSE")
                base_id = clip.data.get('id')
                if base_id:
                    clip_link_info.setdefault(base_id, {})
                    audio_id = "%s-audio" % base_id
                    clip_link_info[base_id]["audio"] = {
                        "node": clipAudioNode,
                        "trackindex": audioTrackNumber or 1,
                        "clipindex": audio_clip_index,
                        "id": audio_id
                    }
                    audio_clip_index += 1
                    link_parts = clip_link_info[base_id]
                    if "video" in link_parts:
                        video_info = link_parts["video"]
                        _append_link(clipAudioNode, "audio", audioTrackNumber or 1, link_parts["audio"]["clipindex"], 1, audio_id)
                        _append_link(clipAudioNode, "video", video_info["trackindex"], video_info["clipindex"], 1, video_info["id"])
                        _append_link(video_info["node"], "video", video_info["trackindex"], video_info["clipindex"], 1, video_info["id"])
                        _append_link(video_info["node"], "audio", audioTrackNumber or 1, link_parts["audio"]["clipindex"], 1, audio_id)

    try:
        with open(os.fsencode(file_path), "wb") as file:
            file.write(bytes(xmldoc.toxml(), 'UTF-8'))
        _validate_export(file_path)
    except IOError as inst:
        log.error("Error writing XML export: {}".format(str(inst)))
    finally:
        # Free up DOM memory
        xmldoc.unlink()



"""
 @file
 @brief Keyboard-search support for tree views with a thumbnail first column
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """


class NameColumnKeyboardSearchMixin:
    """Run Qt's incremental keyboard search against the visible name column."""

    keyboard_search_column = 1

    def keyboardSearch(self, search):
        model = self.model()
        current = self.currentIndex()
        if model and model.rowCount() > 0:
            if current.isValid():
                name_index = current.sibling(current.row(), self.keyboard_search_column)
            else:
                name_index = model.index(0, self.keyboard_search_column)
            if name_index.isValid():
                self.setCurrentIndex(name_index)
        super().keyboardSearch(search)
